# -*- coding: utf-8 -*-
import typing

from kiara import Kiara, PipelineController
from kiara.module_config import ModuleConfig
from kiara.rendering.workflow import WorkflowRenderer


class TerminalRenderer(WorkflowRenderer):
    def __init__(
        self,
        config: typing.Union[ModuleConfig, typing.Mapping[str, typing.Any], str],
        workflow_id: typing.Optional[str] = None,
        module_config: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        controller: typing.Optional[PipelineController] = None,
        kiara: typing.Optional[Kiara] = None,
    ):

        super().__init__(
            config=config,
            workflow_id=workflow_id,
            module_config=module_config,
            controller=controller,
            kiara=kiara,
        )



# -*- coding: utf-8 -*-
import typing
from pydantic import BaseModel, Field
from rich import box
from rich.console import RenderableType
from rich.table import Table

from kiara.info import KiaraInfoModel, extract_renderable

if typing.TYPE_CHECKING:
    from kiara import Kiara
    from kiara.metadata.core_models import MetadataModelMetadata


class MetadataModel(KiaraInfoModel):
    """Base class for classes that represent value metadata in kiara."""

    @classmethod
    def get_type_metadata(cls) -> "MetadataModelMetadata":
        """Return all metadata associated with this module type."""

        from kiara.metadata.core_models import MetadataModelMetadata

        return MetadataModelMetadata.from_model_class(model_cls=cls)

    # @classmethod
    # def model_doc(cls) -> str:
    #
    #     return extract_doc_from_cls(cls)
    #
    # @classmethod
    # def model_desc(cls) -> str:
    #     return extract_doc_from_cls(cls, only_first_line=True)

    # @classmethod
    # def from_dicts(cls, *dicts: typing.Mapping[str, typing.Any]):
    #
    #     if not dicts:
    #         return cls()
    #
    #     merged = merge_dicts(*dicts)
    #     return cls.parse_obj(merged)

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        table = Table(show_header=False, box=box.SIMPLE)
        table.add_column("Key", style="i")
        table.add_column("Value")
        for k in self.__fields__.keys():
            if k == "type_name" and config.get("omit_type_name", False):
                continue
            attr = getattr(self, k)
            v = extract_renderable(attr)
            table.add_row(k, v)

        if "operations" in config.keys():
            ids = list(config["operations"].keys())
            table.add_row("operations", "\n".join(ids))

        return table


class MetadataSet(object):
    def __init__(
        self,
        kiara: "Kiara",
        subject: typing.Any,
        **metadata: typing.Mapping[str, typing.Any],
    ):

        self._kiara: "Kiara" = kiara
        self._data: typing.Dict[str, typing.Mapping[str, typing.Any]] = metadata
        self._metadata: typing.Dict[str, MetadataModel] = {}
        self._subject: typing.Any = subject

    def get_metadata(self, metadata_key: str) -> MetadataModel:

        if metadata_key in self._metadata.keys():
            return self._metadata[metadata_key]

        if metadata_key not in self._data.keys():
            raise Exception(f"No metadata for key '{metadata_key}' available.")

        md = self._data[metadata_key]
        schema = self._kiara.metadata_mgmt.all_schemas.get(metadata_key, None)
        if schema is None:
            raise Exception(
                f"No metadata schema for key '{metadata_key}' registered in kiara."
            )

        metadata_model_obj = schema(**md)
        self._metadata[metadata_key] = metadata_model_obj
        return self._metadata[metadata_key]

    def get_schema(self, metadata_key) -> "MetadataModelMetadata":
        pass


class ValueTypeAndDescription(BaseModel):

    description: str = Field(description="The description for the value.")
    type: str = Field(description="The value type.")



# -*- coding: utf-8 -*-
import pytest

import os
import typing

from kiara import Kiara, Pipeline, PipelineModule
from kiara.pipeline.config import PipelineConfig
from kiara.pipeline.structure import PipelineStep, PipelineStructure

from .utils import INVALID_PIPELINES_FOLDER, PIPELINES_FOLDER


def test_workflow_desc_files(pipeline_paths):

    for path in pipeline_paths.values():
        c = PipelineConfig.parse_file(path)
        assert isinstance(c, PipelineConfig)
        assert len(c.steps) > 0
        assert c.steps[0].step_id
        assert c.steps[0].module_type


def test_workflow_obj_attributes(pipeline_configs: typing.Mapping[str, PipelineConfig]):

    logic_1 = pipeline_configs["logic_1"]

    assert len(logic_1.steps) == 1
    assert len(logic_1.input_aliases) == 0
    assert len(logic_1.output_aliases) == 0

    logic_2 = pipeline_configs["logic_2"]

    assert len(logic_2.steps) == 2
    assert len(logic_2.input_aliases) == 3
    assert len(logic_2.output_aliases) == 1

    logic_3 = pipeline_configs["logic_3"]

    assert len(logic_3.steps) == 3
    assert len(logic_3.input_aliases) == 0
    assert len(logic_3.output_aliases) == 0


def test_workflow_obj_creation(pipeline_configs: typing.Mapping[str, PipelineConfig]):

    logic_1 = pipeline_configs["logic_1"]
    c = PipelineModule(id="logic_1", module_config=logic_1)
    assert isinstance(c, PipelineModule)

    assert c.full_id == "logic_1"
    assert len(c.structure.steps) == 1
    assert "and_1" in c.structure.to_details().steps.keys()
    assert isinstance(c.structure.to_details().steps["and_1"].step, PipelineStep)


def test_pipeline_structure_creation(kiara: Kiara):

    pipeline_file = os.path.join(PIPELINES_FOLDER, "logic", "logic_3.json")

    config = PipelineConfig.create_pipeline_config(pipeline_file)

    structure = PipelineStructure(config=config, kiara=kiara)

    for idx, step in enumerate(structure.steps):
        assert isinstance(step, PipelineStep)

    assert idx == 2

    Pipeline(structure=structure)


def test_invalid_pipeline_creation(kiara: Kiara):

    pipeline_file = os.path.join(INVALID_PIPELINES_FOLDER, "logic_4.json")
    config = PipelineConfig.create_pipeline_config(pipeline_file)

    structure = PipelineStructure(config=config, kiara=kiara)

    for idx, step in enumerate(structure.steps):
        assert isinstance(step, PipelineStep)
    assert idx == 2

    with pytest.raises(Exception) as exc_info:
        Pipeline(structure=structure)
    assert "a1" in str(exc_info.value)



# -*- coding: utf-8 -*-
import typing
from pydantic import Field
from rich import box
from rich.console import RenderableType
from rich.table import Table

from kiara import Kiara
from kiara.info import KiaraInfoModel
from kiara.metadata.operation_models import OperationsMetadata
from kiara.module_config import ModuleConfig
from kiara.operations import Operation, OperationType


class OperationsInfo(KiaraInfoModel):
    @classmethod
    def create_all(self, kiara: Kiara) -> typing.Dict[str, "OperationsInfo"]:

        op_types = kiara.operation_mgmt.operation_types
        return {
            op_name: OperationsInfo.create(operations=op_types[op_name])
            for op_name in sorted(op_types.keys())
        }

    @classmethod
    def create(cls, operations: OperationType):

        info = OperationsMetadata.from_operations_class(operations.__class__)
        return OperationsInfo(
            info=info, operation_configs=dict(operations.operation_configs)
        )

    @classmethod
    def create_renderable_from_operations_map(
        cls, op_map: typing.Mapping[str, "OperationsInfo"]
    ):

        table = Table(show_header=False, box=box.SIMPLE)
        table.add_column("Key", style="i")
        table.add_column("Value")
        for op_name, ops in op_map.items():
            table.add_row(op_name, ops)

        return table

    info: OperationsMetadata = Field(
        description="Details about the type of operations contained in this collection."
    )
    operation_configs: typing.Dict[str, Operation] = Field(
        description="All available operation ids and their configurations."
    )

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        return self.info.create_renderable(
            operations=self.operation_configs, omit_type_name=True
        )


class OperationsGroupInfo(KiaraInfoModel):
    @classmethod
    def create(cls, kiara: "Kiara"):

        operation_types = OperationsInfo.create_all(kiara=kiara)
        operation_configs = operation_types.pop("all")

        return OperationsGroupInfo(
            operation_types=operation_types,
            operation_configs=operation_configs.operation_configs,
        )

    operation_types: typing.Dict[str, OperationsInfo] = Field(
        description="The available operation types and their details."
    )
    operation_configs: typing.Dict[str, Operation] = Field(
        description="The available operation ids and module_configs."
    )

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        table = Table(show_header=False, box=box.SIMPLE)
        table.add_column("Key", style="i")
        table.add_column("Value")

        op_map_table = OperationsInfo.create_renderable_from_operations_map(
            self.operation_types
        )
        table.add_row("operation types", op_map_table)
        configs = ModuleConfig.create_renderable_from_module_instance_configs(
            self.operation_configs
        )
        table.add_row("operations", configs)
        return table



# -*- coding: utf-8 -*-
"""Pipeline-related subcommands for the cli."""

import click
import os.path
import sys

from kiara.info.pipelines import PipelineModuleInfo
from kiara.utils.output import rich_print


@click.group()
@click.pass_context
def pipeline(ctx):
    """Pipeline-related sub-commands."""


@pipeline.command()
@click.argument("pipeline-type", nargs=1)
@click.pass_context
def explain(ctx, pipeline_type: str):
    """Print details about pipeline inputs, outputs, and overall structure."""

    kiara_obj = ctx.obj["kiara"]

    if os.path.isfile(pipeline_type):
        pipeline_type = kiara_obj.register_pipeline_description(
            pipeline_type, raise_exception=True
        )

    m_cls = kiara_obj.get_module_class(pipeline_type)
    if not m_cls.is_pipeline():
        rich_print()
        rich_print(f"Module '{pipeline_type}' is not a pipeline-type module.")
        sys.exit(1)

    info = PipelineModuleInfo.from_type_name(pipeline_type, kiara=kiara_obj)
    print()
    kiara_obj.explain(info.structure)


@pipeline.command()
@click.argument("pipeline-type", nargs=1)
@click.pass_context
def explain_steps(ctx, pipeline_type: str):
    """List all steps of a pipeline."""

    kiara_obj = ctx.obj["kiara"]

    if os.path.isfile(pipeline_type):
        pipeline_type = kiara_obj.register_pipeline_description(
            pipeline_type, raise_exception=True
        )

    m_cls = kiara_obj.get_module_class(pipeline_type)
    if not m_cls.is_pipeline():
        rich_print()
        rich_print(f"Module '{pipeline_type}' is not a pipeline-type module.")
        sys.exit(1)

    info = PipelineModuleInfo.from_type_name(
        module_type_name=pipeline_type, kiara=kiara_obj
    )
    print()
    st_info = info.structure.steps_info
    rich_print(st_info)


@pipeline.command()
@click.argument("pipeline-type", nargs=1)
@click.pass_context
def execution_graph(ctx, pipeline_type: str):
    """Print the execution graph for a pipeline structure."""

    kiara_obj = ctx.obj["kiara"]

    if os.path.isfile(pipeline_type):
        pipeline_type = kiara_obj.register_pipeline_description(
            pipeline_type, raise_exception=True
        )

    m_cls = kiara_obj.get_module_class(pipeline_type)
    if not m_cls.is_pipeline():
        rich_print()
        rich_print(f"Module '{pipeline_type}' is not a pipeline-type module.")
        sys.exit(1)

    info = PipelineModuleInfo.from_type_name(pipeline_type, kiara=kiara_obj)
    info.print_execution_graph()


@pipeline.command()
@click.argument("pipeline-type", nargs=1)
@click.option(
    "--full",
    "-f",
    is_flag=True,
    help="Display full data-flow graph, incl. intermediate input/output connections.",
)
@click.pass_context
def data_flow_graph(ctx, pipeline_type: str, full: bool):
    """Print the data flow graph for a pipeline structure."""

    kiara_obj = ctx.obj["kiara"]
    if os.path.isfile(pipeline_type):
        pipeline_type = kiara_obj.register_pipeline_description(
            pipeline_type, raise_exception=True
        )

    m_cls = kiara_obj.get_module_class(pipeline_type)
    if not m_cls.is_pipeline():
        rich_print()
        rich_print(f"Module '{pipeline_type}' is not a pipeline-type module.")
        sys.exit(1)

    info = PipelineModuleInfo.from_type_name(pipeline_type, kiara=kiara_obj)

    info.print_data_flow_graph(simplified=not full)



# -*- coding: utf-8 -*-
import typing
from pydantic import Field

from kiara.data import ValueSet
from kiara.exceptions import KiaraProcessingException
from kiara.module import KiaraModule
from kiara.module_config import ModuleTypeConfigSchema

if typing.TYPE_CHECKING:
    from kiara.data.values import ValueSchema


class SaveValueModuleConfig(ModuleTypeConfigSchema):

    value_type: str = Field(description="The type of the value to save.")


class SaveValueModule(KiaraModule):
    """Save a value into the kiara data store."""

    _type_name = "value.save"
    _config_cls = SaveValueModuleConfig

    def create_input_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union["ValueSchema", typing.Mapping[str, typing.Any]]
    ]:

        return {
            "value_item": {
                "type": self.get_config_value("value_type"),
                "doc": "The value to save.",
            },
            "aliases": {
                "type": "list",
                "doc": "A list of aliases to link to the saved value id.",
                "optional": True,
            },
        }

    def create_output_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union["ValueSchema", typing.Mapping[str, typing.Any]]
    ]:
        return {
            "value_id": {
                "type": "string",
                "doc": f"The id of the saved {self.get_config_value('value_type')} data-item.",
            }
        }

    def process(self, inputs: ValueSet, outputs: ValueSet) -> None:

        value = inputs.get_value_obj("value_item")
        aliases = inputs.get_value_data("aliases")

        assert value.type_name == self.get_config_value("value_type")

        value = value.save(aliases=aliases)

        outputs.set_value("value_id", value.id)


class LoadValueModule(KiaraModule):
    """Load a value from the kiara data store."""

    _type_name = "value.load"
    _config_cls = SaveValueModuleConfig

    def create_input_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union["ValueSchema", typing.Mapping[str, typing.Any]]
    ]:

        return {
            "value_id": {
                "type": "string",
                "doc": "The id or alias of the saved value you want to load.",
            }
        }

    def create_output_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union["ValueSchema", typing.Mapping[str, typing.Any]]
    ]:

        return {
            "value_item": {
                "type": self.get_config_value("value_type"),
                "doc": "The loaded value.",
            }
        }

    def process(self, inputs: ValueSet, outputs: ValueSet) -> None:

        value_id: str = inputs.get_value_data("value_id")  # type: ignore

        value = self._kiara.data_store.get_value_obj(value_id)
        if value is None:
            raise KiaraProcessingException(f"Can't find value for value id: {value_id}")
        # TODO: make this so we don't have to actually load the data, but can use a reference
        outputs.set_value("value_item", value.get_value_data())



# -*- coding: utf-8 -*-

from click.testing import CliRunner

from kiara.interfaces.cli import cli


def test_metadata_subcommand():

    runner = CliRunner()
    result = runner.invoke(cli, "metadata")
    assert result.exit_code == 0
    assert "Print details for a specific" in result.stdout


def test_metadata_list_subcommand():

    runner = CliRunner()
    result = runner.invoke(cli, "metadata list")

    assert result.exit_code == 0
    assert "table" in result.stdout
    assert "file_bundle" in result.stdout


def test_metadata_explain_subcommand():

    runner = CliRunner()
    result = runner.invoke(cli, "metadata explain table")

    assert result.exit_code == 0
    assert "'table'" in result.stdout
    assert "column_names" in result.stdout



# -*- coding: utf-8 -*-
# %%


from kiara.interfaces.python_api import Step
from kiara.utils.output import rich_print

step_read_files_in_folder = Step(
    "onboarding.folder.import", step_id="read_files_in_folder"
)


step_create_table_from_csvs_config = {
    "columns": ["id", "rel_path", "file_name", "content"]
}
step_create_table_from_csvs = Step(
    "table.from_file_bundle",
    module_config=step_create_table_from_csvs_config,
    step_id="create_table_from_csvs",
)


step_extract_date_from_file_name_config = {"module_type": "date.extract_from_string"}
step_extract_date_from_file_name = Step(
    "array.map",
    module_config=step_extract_date_from_file_name_config,
    step_id="extract_date_from_file_name",
)


step_extract_ref_from_file_name_config = {
    "module_type": "string.match_regex",
    "module_config": {
        "regex": "(\\w+\\d+)_\\d{4}-\\d{2}-\\d{2}_",
        "only_first_match": True,
    },
}
step_extract_ref_from_file_name = Step(
    "array.map",
    module_config=step_extract_ref_from_file_name_config,
    step_id="extract_ref_from_file_name",
)


step_lookup_publication_name_config = {
    "module_type": "string.replace",
    "module_config": {
        "replacement_map": {
            "sn85066408": "L\\'Italia",
            "2012271201": "Cronaca Sovversiva",
        }
    },
}
step_lookup_publication_name = Step(
    "array.map",
    module_config=step_lookup_publication_name_config,
    step_id="lookup_publication_name",
)


step_create_date_range_filter_config = {
    "module_type": "date.range_check",
    "input_name": "date",
}
step_create_date_range_filter = Step(
    "array.map",
    module_config=step_create_date_range_filter_config,
    step_id="create_date_range_filter",
)


step_merged_table = Step("table.merge", step_id="merged_table")


step_filtered_table = Step("table.filter.with_mask", step_id="filtered_table")


step_tokenize_text_corpus_config = {
    "module_type": "language.tokens.tokenize_text",
    "input_name": "text",
}
step_tokenize_text_corpus = Step(
    "array.map",
    module_config=step_tokenize_text_corpus_config,
    step_id="tokenize_text_corpus",
)


step_remove_stopwords = Step(
    "language.tokens.remove_stopwords", step_id="remove_stopwords"
)


step_lemmatize_corpus = Step(
    "language.lemmatize.tokens_array", step_id="lemmatize_corpus"
)


step_generate_lda = Step("language.lda.LDA", step_id="generate_lda")

step_create_table_from_csvs.input.files = step_read_files_in_folder.output.file_bundle


step_extract_date_from_file_name.input.array = (
    step_create_table_from_csvs.output.table.file_name
)


step_extract_ref_from_file_name.input.array = (
    step_create_table_from_csvs.output.table.file_name
)


step_lookup_publication_name.input.array = step_extract_ref_from_file_name.output.array


step_create_date_range_filter.input.array = (
    step_extract_date_from_file_name.output.array
)


step_merged_table.input.sources = [
    step_create_table_from_csvs.output.table,
    step_extract_date_from_file_name.output.array,
    step_extract_ref_from_file_name.output.array,
    step_lookup_publication_name.output.array,
]


step_filtered_table.input.table = step_merged_table.output.table

step_filtered_table.input.mask = step_create_date_range_filter.output.array


step_tokenize_text_corpus.input.array = step_filtered_table.output.table.content


step_remove_stopwords.input.token_lists = step_tokenize_text_corpus.output.array


step_lemmatize_corpus.input.tokens_array = step_remove_stopwords.output.token_list


step_generate_lda.input.tokens_array = step_lemmatize_corpus.output.tokens_array


workflow = step_generate_lda.workflow

step_read_files_in_folder.input.path = (
    "/home/markus/projects/dharpa/notebooks/TopicModelling/data_tm_workflow"
)
step_create_date_range_filter.input.earliest = "1919-01-01"
step_create_date_range_filter.input.latest = "2000-01-01"
step_remove_stopwords.input.languages = ["italian", "german"]
step_generate_lda.input.compute_coherence = True
workflow.process()
rich_print(workflow)



# -*- coding: utf-8 -*-
import mkdocs_gen_files
import os
import typing

from kiara import Kiara

kiara = Kiara.instance()
kiara.available_module_types


def class_namespace(cls: typing.Type):

    module = cls.__module__
    if module is None or module == str.__class__.__module__:
        return cls.__name__
    else:
        return module + "." + cls.__name__


page_file_path = os.path.join("development", "modules_list.md")
page_content = """# Available module types

This page contains a list of all available *Kiara* module types, and their details.

!!! note
The formatting here will be improved later on, for now this should be enough to get the important details of each module type.

"""


for module_type in kiara.available_module_types:

    if module_type == "pipeline":
        continue

    page_content = page_content + f"## ``{module_type}``\n\n"
    page_content = (
        page_content + "```\n{{ get_module_info('" + module_type + "') }}\n```\n\n"
    )

with mkdocs_gen_files.open(page_file_path, "w") as f:
    f.write(page_content)



# -*- coding: utf-8 -*-

"""Main module."""
import logging
import os
import typing
import zmq
from pathlib import Path
from threading import Thread
from zmq.devices import ThreadDevice

from kiara.config import KiaraConfig
from kiara.data import Value
from kiara.data.registry import DataRegistry, InMemoryDataRegistry
from kiara.data.registry.store import LocalDataStore
from kiara.data.types import ValueType
from kiara.data.types.type_mgmt import TypeMgmt
from kiara.data.values.value_set import ValueSet
from kiara.interfaces import get_console
from kiara.metadata.mgmt import MetadataMgmt
from kiara.metadata.module_models import KiaraModuleTypeMetadata
from kiara.module_config import ModuleConfig
from kiara.module_mgmt import ModuleManager
from kiara.module_mgmt.merged import MergedModuleManager
from kiara.operations import Operation, OperationMgmt
from kiara.pipeline.config import PipelineConfig
from kiara.pipeline.controller import PipelineController
from kiara.pipeline.pipeline import Pipeline
from kiara.processing import Job
from kiara.processing.processor import ModuleProcessor
from kiara.utils import (
    create_valid_identifier,
    get_auto_workflow_alias,
    get_data_from_file,
    is_debug,
)
from kiara.workflow.kiara_workflow import KiaraWorkflow

if typing.TYPE_CHECKING:
    from kiara.module import KiaraModule
    from kiara.operations.pretty_print import PrettyPrintOperationType

log = logging.getLogger("kiara")


def explain(item: typing.Any):
    """Pretty print information about an item on the terminal."""

    if isinstance(item, type):
        from kiara.module import KiaraModule

        if issubclass(item, KiaraModule):
            item = KiaraModuleTypeMetadata.from_module_class(module_cls=item)

    elif isinstance(item, Value):
        item.get_metadata()

    console = get_console()
    console.print(item)


class Kiara(object):
    """The core context of a kiara session.

    The `Kiara` object holds all information related to the current environment the user does works in. This includes:

      - available modules, operations & pipelines
      - available value types
      - available metadata schemas
      - available data items
      - available controller and processor types
      - misc. configuration options

    It's possible to use *kiara* without ever manually touching the 'Kiara' class, by default all relevant classes and functions
    will use a default instance of this class (available via the `Kiara.instance()` method.

    The Kiara class is highly dependent on the Python environment it lives in, because it auto-discovers available sub-classes
    of its building blocks (modules, value types, etc.). So, you can't assume that, for example, a pipeline you create
    will work the same way (or at all) in a different environment. *kiara* will always be able to tell you all the details
    of this environment, though, and it will attach those details to things like data, so there is always a record of
    how something was created, and in which environment.
    """

    _instance = None

    @classmethod
    def instance(cls):
        """The default *kiara* context. In most cases, it's recommended you create and manage your own, though."""

        if cls._instance is None:
            cls._instance = Kiara()
        return cls._instance

    def __init__(self, config: typing.Optional[KiaraConfig] = None):

        if not config:
            config = KiaraConfig()

        self._config: KiaraConfig = config

        # self._zmq_context: Context = Context.instance()

        self._operation_mgmt = OperationMgmt(kiara=self)
        self._metadata_mgmt = MetadataMgmt(kiara=self)

        self._data_store = LocalDataStore(kiara=self, base_path=config.data_store)

        # self.start_zmq_device()
        # self.start_log_thread()
        self._default_processor: ModuleProcessor = ModuleProcessor.from_config(
            config.default_processor, kiara=self
        )

        self._type_mgmt_obj: TypeMgmt = TypeMgmt(self)

        self._data_registry: InMemoryDataRegistry = InMemoryDataRegistry(self)
        self._module_mgr: MergedModuleManager = MergedModuleManager(
            config.module_managers, ignore_errors=self._config.ignore_errors
        )

    @property
    def config(self) -> KiaraConfig:
        """The configuration of this *kiara* environment."""
        return self._config

    @property
    def default_processor(self) -> "ModuleProcessor":
        """The default module processor that will be used in this environment, unless otherwise specified."""

        return self._default_processor

    def start_zmq_device(self):

        pd = ThreadDevice(zmq.QUEUE, zmq.SUB, zmq.PUB)
        pd.bind_in("inproc://kiara_in")
        pd.bind_out("inproc://kiara_out")
        pd.setsockopt_in(zmq.SUBSCRIBE, b"")
        pd.setsockopt_in(zmq.IDENTITY, b"SUB")
        pd.setsockopt_out(zmq.IDENTITY, b"PUB")
        pd.start()

    def start_log_thread(self):
        def log_messages():
            socket = self._zmq_context.socket(zmq.SUB)
            socket.setsockopt_string(zmq.SUBSCRIBE, "")
            socket.connect("inproc://kiara_out")

            debug = is_debug()

            while True:
                message = socket.recv()
                topic, details = message.decode().split(" ", maxsplit=1)
                try:
                    job = Job.parse_raw(details)
                    if debug:
                        print(f"{topic}: {job.pipeline_name}.{job.step_id}")
                    else:
                        log.debug(f"{topic}: {job.pipeline_name}.{job.step_id}")

                except Exception as e:
                    if debug:
                        import traceback

                        traceback.print_exception()
                    else:
                        log.debug(e)

        t = Thread(target=log_messages, daemon=True)
        t.start()

    def explain(self, item: typing.Any):
        explain(item)

    @property
    def type_mgmt(self) -> TypeMgmt:

        return self._type_mgmt_obj

    @property
    def data_store(self) -> LocalDataStore:
        return self._data_store

    @property
    def module_mgmt(self) -> MergedModuleManager:
        return self._module_mgr

    @property
    def metadata_mgmt(self) -> MetadataMgmt:
        return self._metadata_mgmt

    @property
    def value_types(self) -> typing.Mapping[str, typing.Type[ValueType]]:
        return self.type_mgmt.value_types

    @property
    def value_type_names(self) -> typing.List[str]:
        return self.type_mgmt.value_type_names

    def determine_type(self, data: typing.Any) -> typing.Optional[ValueType]:

        raise NotImplementedError()

        # return self.type_mgmt.determine_type(data)

    def get_value_type_cls(self, type_name: str) -> typing.Type[ValueType]:

        return self.type_mgmt.get_value_type_cls(type_name=type_name)

    def get_value(self, value_id: str) -> Value:

        if not isinstance(value_id, str):
            raise TypeError(
                f"Invalid type '{type(value_id)}' for value id, must be a string."
            )

        if value_id.startswith("value:"):
            value_id = value_id[6:]

        if not value_id:
            raise Exception("No value id provided.")

        try:
            value = self.data_registry.get_value_obj(value_id)
            if value is None:
                raise Exception(f"No value with id: {value_id}")
            return value
        except Exception:
            pass

        try:
            value = self.data_store.get_value_obj(value_id)
            if value is None:
                raise Exception(f"No value with id: {value_id}")
            return value
        except Exception:
            pass

        raise Exception(f"No value registered for id: {value_id}")

    def add_module_manager(self, module_manager: ModuleManager):

        self._module_mgr.add_module_manager(module_manager)
        self._type_mgmt_obj.invalidate_types()

    @property
    def data_registry(self) -> DataRegistry:
        return self._data_registry

    @property
    def operation_mgmt(self) -> OperationMgmt:
        return self._operation_mgmt

    def get_module_class(self, module_type: str) -> typing.Type["KiaraModule"]:
        return self._module_mgr.get_module_class(module_type=module_type)

    # def get_module_info(self, module_type: str) -> "ModuleInfo":
    #
    #     if module_type not in self.available_module_types:
    #         raise ValueError(f"Module type '{module_type}' not available.")
    #
    #     if module_type in self.available_pipeline_module_types:
    #         from kiara.pipeline.module import PipelineModuleInfo
    #
    #         info = PipelineModuleInfo(module_type=module_type, _kiara=self)  # type: ignore
    #         return info
    #     else:
    #         from kiara.module import ModuleInfo
    #
    #         info = ModuleInfo.from_module_cls(module_cls=module_type)
    #         return info

    @property
    def available_module_types(self) -> typing.List[str]:
        """Return the names of all available modules"""
        return self._module_mgr.available_module_types

    @property
    def available_non_pipeline_module_types(self) -> typing.List[str]:
        """Return the names of all available pipeline-type modules."""

        return self._module_mgr.available_non_pipeline_module_types

    @property
    def available_pipeline_module_types(self) -> typing.List[str]:
        """Return the names of all available pipeline-type modules."""

        return self._module_mgr.available_pipeline_module_types

    @property
    def available_operation_ids(self) -> typing.List[str]:
        return self._operation_mgmt.operation_ids

    def is_pipeline_module(self, module_type: str):

        return self._module_mgr.is_pipeline_module(module_type=module_type)

    def register_pipeline_description(
        self,
        data: typing.Union[Path, str, typing.Mapping[str, typing.Any]],
        module_type_name: typing.Optional[str] = None,
        namespace: typing.Optional[str] = None,
        raise_exception: bool = False,
    ) -> typing.Optional[str]:

        return self._module_mgr.register_pipeline_description(
            data=data,
            module_type_name=module_type_name,
            namespace=namespace,
            raise_exception=raise_exception,
        )

    def create_module(
        self,
        module_type: str,
        module_config: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        id: typing.Optional[str] = None,
        parent_id: typing.Optional[str] = None,
    ) -> "KiaraModule":
        """Create a module instance.

        The 'module_type' argument can be a module type id, or an operation id. In case of the latter, the 'real' module_type
        and module_config will be resolved via the operation management.

        Arguments:
            module_type: the module type- or operation-id
            module_config: the module instance configuration (must be empty in case of the module_type being an operation id
            id: the id of this module, only relevant if the module is part of a pipeline, in which case it must be unique within the pipeline
            parent_id: a reference to the pipeline that contains this module (if applicable)

        Returns:
            The instantiated module object.
        """

        if module_type == "pipeline":
            from kiara import PipelineModule

            module_cls = PipelineModule
            module = module_cls(
                id=id, parent_id=parent_id, module_config=module_config, kiara=self
            )
            return module

        elif (
            module_type not in self.available_module_types
            and module_type in self.operation_mgmt.operation_ids
        ):
            op = self.operation_mgmt.get_operation(module_type)
            assert op is not None
            module_type = op.module_type
            op_config = op.module_config
            if not module_config:
                _module_config: typing.Optional[
                    typing.Mapping[str, typing.Any]
                ] = op_config
            else:
                _module_config = dict(op_config)
                _module_config.update(module_config)
        else:
            _module_config = module_config

        return self._module_mgr.create_module(
            kiara=self,
            id=id,
            module_type=module_type,
            module_config=_module_config,
            parent_id=parent_id,
        )

    def get_module_doc(
        self,
        module_type: str,
        module_config: typing.Optional[typing.Mapping[str, typing.Any]] = None,
    ):

        m = self.create_module(module_type=module_type, module_config=module_config)
        return m.module_instance_doc

    def get_operation(self, operation_id: str) -> Operation:

        op = self.operation_mgmt.profiles.get(operation_id, None)
        if op is None:
            raise Exception(f"No operation with id '{operation_id}' available.")

        return op

    def run(
        self,
        module_type: typing.Union[str, Operation],
        module_config: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        inputs: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        output_name: typing.Optional[str] = None,
        resolve_result: bool = False,
    ) -> typing.Union[ValueSet, Value, typing.Any]:

        if isinstance(module_type, str):
            if module_type in self.available_module_types:
                module = self.create_module(
                    module_type=module_type, module_config=module_config
                )
            elif module_type in self.operation_mgmt.profiles.keys():
                if module_config:
                    raise NotImplementedError()
                op = self.operation_mgmt.profiles[module_type]
                module = op.module
            else:
                raise Exception(
                    f"Can't run operation: invalid module type '{module_type}'"
                )
        elif isinstance(module_type, Operation):
            if module_config:
                raise NotImplementedError()
            module = module_type.module
        else:
            raise Exception(f"Invalid class for module_type: {type(module_type)}")

        return self.run_module(
            module=module,
            inputs=inputs,
            output_name=output_name,
            resolve_result=resolve_result,
        )

    def run_module(
        self,
        module: "KiaraModule",
        inputs: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        output_name: typing.Optional[str] = None,
        resolve_result: bool = False,
    ):

        if inputs is None:
            inputs = {}
        result = module.run(**inputs)
        if output_name is not None:
            v = result.get_value_obj(output_name)
            if resolve_result:
                return v.get_value_data()
            else:
                return v
        else:
            if resolve_result:
                return result.get_all_value_data()
            else:
                return result

    def create_pipeline(
        self,
        config: typing.Union[ModuleConfig, typing.Mapping[str, typing.Any], str],
        controller: typing.Optional[PipelineController] = None,
    ) -> Pipeline:

        if isinstance(config, typing.Mapping):
            pipeline_config: PipelineConfig = PipelineConfig(**config)

        elif isinstance(config, str):
            if config == "pipeline":
                raise Exception(
                    "Can't create pipeline from 'pipeline' module type without further configuration."
                )

            if config in self.available_module_types:
                config_data = {
                    "steps": [
                        {
                            "module_type": config,
                            "step_id": create_valid_identifier(config),
                        }
                    ]
                }
                pipeline_config = PipelineConfig(**config_data)
            elif os.path.isfile(os.path.expanduser(config)):
                path = os.path.expanduser(config)
                pipeline_config_data = get_data_from_file(path)
                pipeline_config = PipelineConfig(**pipeline_config_data)
            else:
                raise Exception(
                    f"Can't create pipeline config from string '{config}'. Value must be path to a file, or one of: {', '.join(self.available_module_types)}"
                )
        elif isinstance(config, PipelineConfig):
            pipeline_config = config
        else:
            raise TypeError(
                f"Invalid type '{type(config)}' for pipeline configuration."
            )

        pipeline = pipeline_config.create_pipeline(controller=controller, kiara=self)
        return pipeline

    def create_workflow_from_operation_config(
        self,
        config: "ModuleConfig",
        workflow_id: typing.Optional[str] = None,
        controller: typing.Optional[PipelineController] = None,
    ):

        if not workflow_id:
            workflow_id = get_auto_workflow_alias(
                config.module_type, use_incremental_ids=True
            )

        workflow = KiaraWorkflow(
            workflow_id=workflow_id,
            config=config,
            controller=controller,
            kiara=self,
        )
        return workflow

    def create_workflow(
        self,
        config: typing.Union[ModuleConfig, typing.Mapping[str, typing.Any], str],
        workflow_id: typing.Optional[str] = None,
        module_config: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        controller: typing.Optional[PipelineController] = None,
    ) -> KiaraWorkflow:

        _config = ModuleConfig.create_module_config(
            config=config, module_config=module_config, kiara=self
        )
        return self.create_workflow_from_operation_config(
            config=_config, workflow_id=workflow_id, controller=controller
        )

    def pretty_print(
        self,
        value: Value,
        target_type: str = "renderables",
        print_config: typing.Optional[typing.Mapping[str, typing.Any]] = None,
    ) -> typing.Any:

        pretty_print_ops: PrettyPrintOperationType = self.operation_mgmt.get_operations("pretty_print")  # type: ignore

        return pretty_print_ops.pretty_print(
            value=value, target_type=target_type, print_config=print_config
        )



# -*- coding: utf-8 -*-
import pytest

from kiara import Kiara


def test_simple_module_exec(kiara: Kiara):

    and_mod = kiara.create_module("logic.and")

    # boolean_schema = ValueSchema(type="boolean")

    inputs = {"a": None, "b": True}
    with pytest.raises(Exception) as e:
        and_mod.run(**inputs)

    assert "Inputs not valid" in str(e.value)

    # inp_a = Value(value_data=True, value_schema=boolean_schema, kiara=kiara)
    # inp_b = Value(value_data=True, value_schema=boolean_schema, kiara=kiara)
    inputs = {"a": True, "b": True}
    outputs = and_mod.run(**inputs)
    assert outputs.get_value_data("y") is True

    inputs = {"a": False, "b": True}
    outputs = and_mod.run(**inputs)
    assert outputs.get_value_data("y") is False



# -*- coding: utf-8 -*-

"""Module-related configuration models for the *Kiara* package."""

import deepdiff
import os
import typing
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator
from rich import box
from rich.console import Console, ConsoleOptions, RenderableType, RenderResult
from rich.syntax import Syntax
from rich.table import Table

from kiara.defaults import DEFAULT_NO_DESC_VALUE
from kiara.info import KiaraInfoModel
from kiara.metadata.core_models import DocumentationMetadataModel
from kiara.utils import get_data_from_file

if typing.TYPE_CHECKING:
    from kiara.kiara import Kiara
    from kiara.module import KiaraModule
    from kiara.pipeline.config import PipelineConfig


class ModuleTypeConfigSchema(BaseModel):
    """Base class that describes the configuration a [``KiaraModule``][kiara.module.KiaraModule] class accepts.

    This is stored in the ``_config_cls`` class attribute in each ``KiaraModule`` class.

    There are two config options every ``KiaraModule`` supports:

     - ``constants``, and
     - ``defaults``

     Constants are pre-set inputs, and users can't change them and an error is thrown if they try. Defaults are default
     values that override the schema defaults, and those can be overwritten by users. If both a constant and a default
     value is set for an input field, an error is thrown.
    """

    @classmethod
    def requires_config(cls) -> bool:
        """Return whether this class can be used as-is, or requires configuration before an instance can be created."""

        for field_name, field in cls.__fields__.items():
            if field.required and field.default is None:
                return True
        return False

    _config_hash: str = PrivateAttr(default=None)
    constants: typing.Dict[str, typing.Any] = Field(
        default_factory=dict, description="Value constants for this module."
    )
    defaults: typing.Dict[str, typing.Any] = Field(
        default_factory=dict, description="Value defaults for this module."
    )

    class Config:
        extra = Extra.forbid
        validate_assignment = True

    def get(self, key: str) -> typing.Any:
        """Get the value for the specified configuation key."""

        if key not in self.__fields__:
            raise Exception(
                f"No config value '{key}' in module config class '{self.__class__.__name__}'."
            )

        return getattr(self, key)

    @property
    def config_hash(self):

        if self._config_hash is None:
            _d = self.dict()
            hashes = deepdiff.DeepHash(_d)
            self._config_hash = hashes[_d]
        return self._config_hash

    def __eq__(self, other):

        if self.__class__ != other.__class__:
            return False

        return self.dict() == other.dict()

    def __hash__(self):

        return hash(self.config_hash)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        my_table = Table(box=box.MINIMAL, show_header=False)
        my_table.add_column("Field name", style="i")
        my_table.add_column("Value")
        for field in self.__fields__:
            my_table.add_row(field, getattr(self, field))

        yield my_table


def parse_and_create_module_config(
    config: typing.Union[
        "ModuleConfig",
        typing.Mapping,
        str,
        typing.Type["KiaraModule"],
        "PipelineConfig",
    ],
    module_config: typing.Union[None, typing.Mapping[str, typing.Any]] = None,
    kiara: typing.Optional["Kiara"] = None,
) -> "ModuleConfig":
    """Utility method to create a `ModuleConfig` object from different input types.

    Arguments:
        config: the 'main' configuation
        module_config: if the 'main' configuration value is an (unconfigured) module type, this argument can contain the configuration for the module instance
        kiara: the *kiara* context
    """

    if kiara is None:
        from kiara.kiara import Kiara

        kiara = Kiara.instance()

    if isinstance(config, type):
        from kiara.module import KiaraModule

        if issubclass(config, KiaraModule):
            # this basically makes sure the class is augmented with the '_module_type_id` attribute
            if not hasattr(config, "_module_type_id"):
                match = False
                for mod_name in kiara.available_module_types:
                    mod_cls = kiara.get_module_class(mod_name)
                    if mod_cls == config:
                        match = True
                        break
                if not match:
                    raise Exception(f"Class '{config}' not a valid kiara module.")

            config = config._module_type_id  # type: ignore
        else:
            raise TypeError(f"Invalid type '{type(config)}' for module configuration.")

    if isinstance(config, typing.Mapping):

        if module_config:
            raise NotImplementedError()

        module_instance_config: ModuleConfig = ModuleConfig(**config)

    elif isinstance(config, str):
        if config == "pipeline":
            if not module_config:
                raise Exception(
                    "Can't create module from 'pipeline' module type without further configuration."
                )
            elif isinstance(module_config, typing.Mapping):
                module_instance_config = ModuleConfig(
                    module_type="pipeline", module_config=module_config
                )
            else:
                from kiara.pipeline.config import PipelineConfig

                if isinstance(config, PipelineConfig):
                    module_instance_config = ModuleConfig(
                        module_type="pipeline", module_config=config.dict()
                    )
                else:
                    raise Exception(
                        f"Can't create module config, invalid type for 'module_config' parameter: {type(module_config)}"
                    )

        elif config in kiara.available_module_types:
            if module_config:
                module_instance_config = ModuleConfig(
                    module_type=config, module_config=module_config
                )
            else:
                module_instance_config = ModuleConfig(module_type=config)
        elif config in kiara.operation_mgmt.profiles.keys():
            module_instance_config = kiara.operation_mgmt.profiles[config]
        elif os.path.isfile(os.path.expanduser(config)):
            path = os.path.expanduser(config)
            module_config_data = get_data_from_file(path)

            if module_config:
                raise NotImplementedError()

            if not isinstance(module_config_data, typing.Mapping):
                raise Exception(
                    f"Invalid module/pipeline config, must be a mapping type: {module_config_data}"
                )

            if "steps" in module_config_data.keys():
                module_type = "pipeline"
                module_instance_config = ModuleConfig(
                    module_type=module_type, module_config=module_config_data
                )
            elif "module_type" in module_config_data.keys():
                module_instance_config = ModuleConfig(**module_config_data)
        else:
            raise Exception(
                f"Can't create module config from string '{config}'. Value must be path to a file, or one of: {', '.join(kiara.available_module_types)}"
            )
    elif isinstance(config, ModuleConfig):
        module_instance_config = config
        if module_config:
            raise NotImplementedError()
    else:

        from kiara.pipeline.config import PipelineConfig

        if isinstance(config, PipelineConfig):
            module_instance_config = ModuleConfig(
                module_type="pipeline", module_config=config.dict()
            )
        else:
            raise TypeError(f"Invalid type '{type(config)}' for module configuration.")

    return module_instance_config


class ModuleConfig(KiaraInfoModel):
    """A class to hold the type and configuration for a module instance."""

    @classmethod
    def create_module_config(
        cls,
        config: typing.Union[
            "ModuleConfig",
            typing.Mapping,
            str,
            typing.Type["KiaraModule"],
            "PipelineConfig",
        ],
        module_config: typing.Union[None, typing.Mapping[str, typing.Any]] = None,
        kiara: typing.Optional["Kiara"] = None,
    ) -> "ModuleConfig":

        conf = parse_and_create_module_config(
            config=config, module_config=module_config, kiara=kiara
        )
        return conf

    @classmethod
    def create_renderable_from_module_instance_configs(
        cls,
        configs: typing.Mapping[str, "ModuleConfig"],
        **render_config: typing.Any,
    ):
        """Convenience method to create a renderable for this module configuration, to be printed to terminal."""

        table = Table(show_header=False, box=box.SIMPLE)
        table.add_column("Id", style="i", no_wrap=True)
        table.add_column("Description")
        if not render_config.get("omit_config", False):
            table.add_column("Configuration")

        for id, config in configs.items():
            if config.doc:
                doc = config.doc.description
            else:
                doc = DEFAULT_NO_DESC_VALUE

            row: typing.List[RenderableType] = [id, doc]
            if not render_config.get("omit_config", False):
                conf = config.json(exclude={"doc"}, indent=2)
                row.append(Syntax(conf, "json", background_color="default"))
            table.add_row(*row)

        return table

    class Config:
        extra = Extra.forbid
        validate_all = True

    _module: typing.Optional["KiaraModule"] = PrivateAttr(default=None)
    module_type: str = Field(description="The module type.")
    module_config: typing.Dict[str, typing.Any] = Field(
        default_factory=dict, description="The configuration for the module."
    )
    doc: DocumentationMetadataModel = Field(
        description="Documentation for this operation.", default=None
    )

    @validator("doc", pre=True)
    def create_doc(cls, value):
        return DocumentationMetadataModel.create(value)

    def create_module(
        self,
        kiara: typing.Optional["Kiara"] = None,
        module_id: typing.Optional[str] = None,
    ) -> "KiaraModule":
        """Create a module instance from this configuration."""

        if module_id and not isinstance(module_id, str):
            raise TypeError(
                f"Invalid type, module_id must be a string, not: {type(module_id)}"
            )

        if kiara is None:
            from kiara.kiara import Kiara

            kiara = Kiara.instance()

        if self._module is None:
            self._module = kiara.create_module(
                id=module_id,
                module_type=self.module_type,
                module_config=self.module_config,
            )
        return self._module

    def create_renderable(self, **config: typing.Any) -> RenderableType:
        """Create a renderable for this module configuration."""

        conf = Syntax(
            self.json(exclude_none=True, indent=2),
            "json",
            background_color="default",
        )
        return conf


KIARA_CONFIG = typing.TypeVar("KIARA_CONFIG", bound=ModuleTypeConfigSchema)



# -*- coding: utf-8 -*-
import typing
from rich import box
from rich.console import RenderableType
from rich.table import Table

from kiara import Kiara
from kiara.info import KiaraInfoModel
from kiara.metadata.core_models import MetadataModelMetadata


class MetadataModelsInfo(KiaraInfoModel):

    __root__: typing.Dict[str, MetadataModelMetadata]

    @classmethod
    def from_metadata_keys(cls, *metadata_keys: str, kiara: typing.Optional[Kiara]):

        if kiara is None:
            kiara = Kiara.instance()

        models = {}
        invalid = []
        for metadata_key in metadata_keys:
            model = kiara.metadata_mgmt.all_schemas.get(metadata_key, None)
            if model is None:
                invalid.append(metadata_key)
            else:
                models[metadata_key] = MetadataModelMetadata.from_model_class(
                    kiara.metadata_mgmt.all_schemas[metadata_key]
                )

        if invalid:
            raise Exception(
                f"Can't create metadata information, one or several metadata keys are not registered: {', '.join(invalid)}"
            )

        return MetadataModelsInfo(__root__=models)

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        table = Table(show_header=True, box=box.SIMPLE)
        table.add_column("Metadata key", style="i")
        table.add_column("Description")

        for key, model in self.__root__.items():
            table.add_row(key, model.documentation.description)

        return table



# -*- coding: utf-8 -*-

from kiara import Kiara


def test_operation_type_list(kiara: Kiara):

    assert "calculate_hash" in kiara.operation_mgmt.operation_types.keys()

    ch_ops = kiara.operation_mgmt.get_operations("calculate_hash")
    ch_tm = ch_ops.get_type_metadata()
    assert ch_tm.type_name == "calculate_hash"



# -*- coding: utf-8 -*-
"""The 'run' subcommand for the cli."""
import click
import copy
import os
import typing
from pydantic import Field, PrivateAttr
from rich import box
from rich.console import RenderableType
from rich.panel import Panel
from rich.table import Table

from kiara import Kiara
from kiara.data import Value
from kiara.data.values import ValueAlias, ValueInfo
from kiara.info import KiaraInfoModel
from kiara.metadata.module_models import KiaraModuleTypeMetadata
from kiara.metadata.operation_models import OperationsMetadata
from kiara.module import KiaraModule
from kiara.operations import Operation, OperationType
from kiara.utils import log_message
from kiara.utils.output import rich_print


def find_operation_types(
    item: str, kiara: typing.Optional[Kiara] = None
) -> typing.Dict[str, OperationsMetadata]:

    if kiara is None:
        kiara = Kiara.instance()

    all_operations_types = kiara.operation_mgmt.operation_types

    matches = {}
    for operation_name in sorted(all_operations_types.keys()):
        if operation_name == "all":
            continue

        operation_details: OperationType = all_operations_types[operation_name]
        if item in operation_name:
            matches[operation_name] = OperationsMetadata.from_operations_class(
                operation_details.__class__
            )
            continue

        desc = operation_details.get_type_metadata().documentation.description
        if item in desc:
            matches[operation_name] = OperationsMetadata.from_operations_class(
                operation_details.__class__
            )

    return matches


def find_operations(
    item: str,
    search_desc: bool = True,
    search_full_doc: bool = False,
    kiara: typing.Optional[Kiara] = None,
) -> typing.Dict[str, Operation]:

    if kiara is None:
        kiara = Kiara.instance()

    all_operations_types = kiara.operation_mgmt.operation_types

    matches = {}
    for operation_name in all_operations_types.keys():
        if operation_name == "all":
            continue

        operation_details: OperationType = all_operations_types[operation_name]

        for op_id, op_config in sorted(operation_details.operation_configs.items()):

            if item in op_id:
                matches[op_id] = op_config
                continue

            if not search_desc:
                continue

            if search_full_doc:
                desc = op_config.doc.full_doc
            else:
                desc = op_config.doc.description

            if item in desc:
                matches[op_id] = op_config

    return matches


def find_module_types(
    item: str,
    search_desc: bool = True,
    search_full_doc: bool = False,
    kiara: typing.Optional[Kiara] = None,
) -> typing.Dict[str, KiaraModuleTypeMetadata]:

    if kiara is None:
        kiara = Kiara.instance()

    matches = {}
    for module_id in kiara.available_module_types:
        if item in module_id:
            module_cls = kiara.get_module_class(module_id)
            matches[module_id] = KiaraModuleTypeMetadata.from_module_class(module_cls)
            continue

        if not search_desc:
            continue

        module_cls = kiara.get_module_class(module_id)
        if search_full_doc:
            desc = module_cls.get_type_metadata().documentation.full_doc
        else:
            desc = module_cls.get_type_metadata().documentation.description

        if item in desc:
            matches[module_id] = KiaraModuleTypeMetadata.from_module_class(module_cls)

    return matches


def find_values(
    item: str,
    search_ids: bool = True,
    search_aliases: bool = True,
    kiara: typing.Optional[Kiara] = None,
) -> typing.Dict[str, Value]:

    if kiara is None:
        kiara = Kiara.instance()

    matches = {}
    if search_ids:
        for value_id in kiara.data_store.value_ids:
            if item in value_id:
                value_obj = kiara.data_store.get_value_obj(value_id)
                matches[value_id] = ValueInfo.from_value(value_obj)  # type: ignore

    if search_aliases:
        _alias_obj = ValueAlias.from_string(item)
        if _alias_obj.tag or _alias_obj.version:
            match = kiara.data_store.get_value_obj(_alias_obj)
            if match:
                matches[_alias_obj.full_alias] = ValueInfo.from_value(match)
        else:
            for alias in kiara.data_store.alias_names:
                if item in alias:
                    value_slot = kiara.data_store.get_value_slot(alias)
                    assert value_slot
                    value = value_slot.get_latest_value()
                    assert value
                    matches[alias] = ValueInfo.from_value(value)

    return matches


class KiaraEntityMatches(KiaraInfoModel):
    @classmethod
    def search(
        self,
        search_term: str,
        search_ids: bool = True,
        search_aliases: bool = True,
        search_desc: bool = True,
        search_full_doc: bool = False,
        kiara: typing.Optional[Kiara] = None,
    ):

        if kiara is None:
            kiara = Kiara.instance()

        module_types = find_module_types(
            item=search_term,
            search_desc=search_desc,
            search_full_doc=search_full_doc,
            kiara=kiara,
        )
        operation_types = find_operation_types(item=search_term, kiara=kiara)
        operations = find_operations(
            item=search_term,
            search_desc=search_desc,
            search_full_doc=search_full_doc,
            kiara=kiara,
        )
        values = find_values(
            item=search_term,
            search_ids=search_ids,
            search_aliases=search_aliases,
            kiara=kiara,
        )

        return KiaraEntityMatches(
            module_types=module_types,
            operation_types=operation_types,
            operations=operations,
            values=values,
            kiara=kiara,
        )

    def __init__(
        self,
        module_types: typing.Optional[
            typing.Dict[str, typing.Type[KiaraModule]]
        ] = None,
        operation_types: typing.Optional[typing.Dict[str, OperationsMetadata]] = None,
        operations: typing.Optional[typing.Dict[str, Operation]] = None,
        values: typing.Optional[typing.Dict[str, Value]] = None,
        kiara: Kiara = None,
    ):

        if kiara is None:
            kiara = Kiara.instance()

        init_dict: typing.Dict[str, typing.Any] = {}
        if module_types:
            init_dict["module_types"] = module_types
        if operation_types:
            init_dict["operation_types"] = operation_types
        if operations:
            init_dict["operations"] = operations
        if values:
            init_dict["values"] = values

        super().__init__(**init_dict)
        self._kiara = kiara

    _kiara: Kiara = PrivateAttr()

    module_types: typing.Dict[str, KiaraModuleTypeMetadata] = Field(
        description="Matching module types.", default_factory=dict
    )
    operation_types: typing.Dict[str, OperationsMetadata] = Field(
        description="Matching operation types.", default_factory=dict
    )
    operations: typing.Dict[str, Operation] = Field(
        description="Matching operations.", default_factory=dict
    )
    values: typing.Dict[str, ValueInfo] = Field(
        description="Matching values.", default_factory=dict
    )

    @property
    def no_module_types(self) -> int:
        return len(self.module_types)

    @property
    def no_operation_types(self) -> int:
        return len(self.operation_types)

    @property
    def no_operations(self) -> int:
        return len(self.operations)

    @property
    def no_values(self) -> int:
        return len(self.values)

    @property
    def no_results(self) -> int:

        return (
            self.no_module_types
            + self.no_operation_types
            + self.no_operations
            + self.no_values
        )

    def get_single_result(
        self,
    ) -> typing.Optional[typing.Tuple[typing.Type, str, typing.Any]]:

        if self.no_results != 1:
            return None

        if self.module_types:
            match_key = next(iter(self.module_types.keys()))
            match_value: typing.Any = self.module_types[match_key]
            return (KiaraModuleTypeMetadata, match_key, match_value)

        if self.operation_types:
            match_key = next(iter(self.operation_types.keys()))
            match_value = self.operation_types[match_key]
            return (OperationsMetadata, match_key, match_value)

        if self.operations:
            match_key = next(iter(self.operations.keys()))
            match_value = self.operations[match_key]
            return (Operation, match_key, match_value)

        if self.values:
            match_key = next(iter(self.values.keys()))
            match_value = self.values[match_key]
            return (ValueInfo, match_key, match_value)

        raise Exception("No match found. This is a bug.")

    def merge(self, other: "KiaraEntityMatches") -> "KiaraEntityMatches":

        module_types = copy.copy(self.module_types)
        module_types.update(other.module_types)

        operation_types = copy.copy(self.operation_types)
        operation_types.update(other.operation_types)

        operations = copy.copy(self.operations)
        operations.update(other.operations)

        values = copy.copy(self.values)
        values.update(other.values)

        return KiaraEntityMatches(
            module_types=module_types,
            operation_types=operation_types,
            operations=operations,
            values=values,
            kiara=self._kiara,
        )

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        table = Table(show_header=False, box=box.SIMPLE)
        table.add_column("Id")
        table.add_column("Description")

        empty_line = False

        if self.module_types:

            table.add_row()
            table.add_row("[b]Module types[/b]", "")
            table.add_row()
            for module_type, module_type_info in self.module_types.items():
                table.add_row(
                    f"  [i]{module_type}[/i]",
                    module_type_info.documentation.description,
                )

            empty_line = True

        if self.operation_types:
            if empty_line:
                table.add_row()
            table.add_row("[b]Operation types[/b]", "")
            table.add_row()
            for op_type, op_type_info in self.operation_types.items():
                table.add_row(
                    f"  [i]{op_type}[/i]", op_type_info.documentation.description
                )

            empty_line = True

        if self.operations:
            if empty_line:
                table.add_row()
            table.add_row("[b]Operations[/b]", "")
            table.add_row()
            for op_id, op_info in self.operations.items():
                table.add_row(f"  [i]{op_id}[/i]", op_info.doc.description)

            empty_line = True

        if self.values:
            if empty_line:
                table.add_row()
            table.add_row("[b]Values[/b]", "")
            table.add_row()
            for value_id, value in self.values.items():
                table.add_row(f"  [i]{value_id}[/i]", value.value_schema.type)

        return table


@click.command()
@click.argument("search_terms", required=True, nargs=-1, metavar="search_term")
@click.pass_context
def explain(ctx, search_terms):
    """Find and explain any entity in kiara."""

    kiara_obj: Kiara = ctx.obj["kiara"]

    search_ids = True
    search_aliases = True
    search_desc = True
    search_full_doc = False

    result = KiaraEntityMatches(kiara=kiara_obj)

    for search_term in search_terms:

        if os.path.isfile(os.path.realpath(search_term)):
            try:
                search_term = kiara_obj.register_pipeline_description(data=search_term)
            except Exception as e:
                log_message(f"Tried to import '{search_term}' as pipeline, failed: {e}")

        matches = KiaraEntityMatches.search(
            search_term=search_term,
            search_desc=search_desc,
            search_full_doc=search_full_doc,
            search_ids=search_ids,
            search_aliases=search_aliases,
            kiara=kiara_obj,
        )
        result = result.merge(matches)

    if result.no_results == 1:
        print()
        result_type, result_id, result_value = result.get_single_result()
        if result_type == KiaraModuleTypeMetadata:
            title = "Module type"
        elif result_type == OperationsMetadata:
            title = "Operations type"
        elif result_type == Operation:
            title = "Operation"
        elif result_type == ValueInfo:
            title = "Value"
        else:
            raise Exception("Invalid result type. This is a bug.")

        title = f"{title}: [b]{result_id}[/b]"
        rich_print(
            Panel(result_value, title=title, title_align="left", box=box.ROUNDED)
        )
    else:
        rich_print(result)



# -*- coding: utf-8 -*-
import typing
from pydantic import Field

from kiara import KiaraModule
from kiara.data.values import ValueSchema
from kiara.module_config import ModuleTypeConfigSchema


class ValueMergeModuleConfig(ModuleTypeConfigSchema):

    input_schemas: typing.Dict[str, typing.Mapping[str, typing.Any]] = Field(
        description="The schemas for all of the expected inputs."
    )
    output_type: str = Field(description="The result type of the merged value.")


class ValueMergeModule(KiaraModule):
    def create_input_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:

        input_schema_dicts: typing.Mapping[
            str, typing.Mapping[str, typing.Any]
        ] = self.get_config_value("input_schemas")
        if not input_schema_dicts:
            raise Exception("No input schemas provided.")

        input_schemas: typing.Dict[str, ValueSchema] = {}
        for k, v in input_schema_dicts.items():
            input_schemas[k] = ValueSchema(**v)

        return input_schemas

    def create_output_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:

        return {"merged_value": {"type": self.get_config_value("output_type")}}



# -*- coding: utf-8 -*-
import typing
from jinja2 import Environment, FileSystemLoader, Template
from pathlib import Path

from kiara import PipelineStructure


class PipelineRenderer(object):
    def __init__(self, structure: PipelineStructure):

        self._structure: PipelineStructure = structure

    def render_from_path(
        self, path: typing.Union[str, Path], inputs: typing.Mapping[str, typing.Any]
    ):

        if not isinstance(path, Path):
            path = Path(path)

        loader = FileSystemLoader(path.parent)
        env = Environment(loader=loader)

        template = env.get_template(path.name)
        return self.render(template, inputs=inputs)

    def render(self, template: Template, inputs: typing.Mapping[str, typing.Any]):

        rendered = template.render(structure=self._structure, input_values=inputs)
        return rendered



# -*- coding: utf-8 -*-
import abc

from kiara.events import (
    PipelineInputEvent,
    PipelineOutputEvent,
    StepInputEvent,
    StepOutputEvent,
)


class PipelineListener(abc.ABC):
    def step_inputs_changed(self, event: StepInputEvent):
        """Method to override if the implementing controller needs to react to events where one or several step inputs have changed.

        Arguments:
            event: the step input event
        """

    def step_outputs_changed(self, event: StepOutputEvent):
        """Method to override if the implementing controller needs to react to events where one or several step outputs have changed.

        Arguments:
            event: the step output event
        """

    def pipeline_inputs_changed(self, event: PipelineInputEvent):
        """Method to override if the implementing controller needs to react to events where one or several pipeline inputs have changed.

        !!! note
        Whenever pipeline inputs change, the connected step inputs also change and an (extra) event will be fired for those. Which means
        you can choose to only implement the ``step_inputs_changed`` method if you want to. This behaviour might change in the future.

        Arguments:
            event: the pipeline input event
        """

    def pipeline_outputs_changed(self, event: PipelineOutputEvent):
        """Method to override if the implementing controller needs to react to events where one or several pipeline outputs have changed.

        Arguments:
            event: the pipeline output event
        """


# class DebugListener(PipelineListener):
#     def step_inputs_changed(self, event: StepInputEvent):
#
#         pp(event.dict())
#
#     def step_outputs_changed(self, event: StepOutputEvent):
#
#         pp(event.dict())
#
#     def pipeline_inputs_changed(self, event: PipelineInputEvent):
#
#         pp(event.dict())
#
#     def pipeline_outputs_changed(self, event: PipelineOutputEvent):
#
#         pp(event.dict())



# -*- coding: utf-8 -*-
import click
import os
import sys
import typing
from rich import box
from rich.panel import Panel
from rich.table import Table

from kiara import Kiara
from kiara.operations import OperationType
from kiara.utils import log_message
from kiara.utils.output import rich_print


@click.group()
@click.pass_context
def operation(ctx):
    """Metadata-related sub-commands."""


@operation.command(name="list")
@click.option(
    "--by-type", "-t", is_flag=True, help="List the operations by operation type."
)
@click.argument("filter", nargs=-1, required=False)
@click.option(
    "--full-doc", "-d", is_flag=True, help="Display the full doc for all operations."
)
@click.option(
    "--omit-default",
    "-o",
    is_flag=True,
    help="Don't list operations that have no specific operation type associated with them.",
)
@click.pass_context
def list(
    ctx, by_type: bool, filter: typing.Iterable[str], full_doc: bool, omit_default: bool
):

    kiara_obj: Kiara = ctx.obj["kiara"]

    if by_type:
        title = "Operations by type"
        all_operations_types = kiara_obj.operation_mgmt.operation_types

        table = Table(box=box.SIMPLE, show_header=True)
        table.add_column("Type", no_wrap=True, style="b green")
        table.add_column("Id", no_wrap=True)
        table.add_column("Description", no_wrap=False, style="i")

        for operation_name in sorted(all_operations_types.keys()):

            if operation_name == "all":
                continue

            operation_details: OperationType = all_operations_types[operation_name]
            first_line_value = True

            for op_id, op_config in sorted(operation_details.operation_configs.items()):

                if full_doc:
                    desc = op_config.doc.full_doc
                else:
                    desc = op_config.doc.description

                if filter:
                    match = True
                    for f in filter:
                        if (
                            f.lower() not in operation_name.lower()
                            and f.lower() not in op_id.lower()
                            and f.lower() not in desc.lower()
                        ):
                            match = False
                            break
                    if not match:
                        continue

                row = []
                if first_line_value:
                    row.append(operation_name)
                else:
                    row.append("")

                row.append(op_id)
                row.append(desc)

                table.add_row(*row)
                first_line_value = False

    else:
        title = "All operations"
        table = Table(box=box.SIMPLE, show_header=True)
        table.add_column("Id", no_wrap=True, style="b")
        table.add_column("Type(s)", style="green")
        table.add_column("Description", style="i")

        for op_id, config in kiara_obj.operation_mgmt.profiles.items():

            types = kiara_obj.operation_mgmt.get_types_for_id(op_id)

            if omit_default and len(types) == 1:
                continue

            types.remove("all")

            if full_doc:
                desc = config.doc.full_doc
            else:
                desc = config.doc.description
            # desc = config.module_cls.get_type_metadata().documentation.description
            if filter:
                match = True
                for f in filter:
                    if f.lower() not in op_id.lower() and f.lower() not in desc.lower():
                        match = False
                        break
                if match:
                    table.add_row(op_id, ", ".join(types), desc)

            else:
                table.add_row(op_id, ", ".join(types), desc)

    panel = Panel(table, title=title, title_align="left", box=box.ROUNDED)
    rich_print(panel)


@operation.command()
@click.argument("operation_id", nargs=1, required=True)
@click.option(
    "--source",
    "-s",
    help="Show module source code (or pipeline configuration).",
    is_flag=True,
)
@click.pass_context
def explain(ctx, operation_id: str, source: bool):

    kiara_obj: Kiara = ctx.obj["kiara"]

    if os.path.isfile(os.path.realpath(operation_id)):
        try:
            _operation_id = kiara_obj.register_pipeline_description(data=operation_id)
            if _operation_id is None:
                print(
                    f"Unknown error when trying to import '{operation_id}' as pipeline."
                )
                sys.exit(1)
            else:
                operation_id = _operation_id
        except Exception as e:
            log_message(f"Tried to import '{operation_id}' as pipeline, failed: {e}")

    op_config = kiara_obj.operation_mgmt.profiles.get(operation_id)
    if not op_config:
        print()
        print(f"No operation with id '{operation_id}' registered.")
        sys.exit(1)

    print()
    rich_print(
        op_config.create_panel(
            title=f"Operation: [b i]{operation_id}[/b i]", include_src=source
        )
    )



# -*- coding: utf-8 -*-
import typing

if typing.TYPE_CHECKING:
    from kiara import KiaraModule
    from kiara.data.types import ValueType
    from kiara.data.values.value_set import ValueSet


class KiaraException(Exception):
    pass


class KiaraModuleConfigException(Exception):
    def __init__(
        self,
        msg: str,
        module_cls: typing.Type["KiaraModule"],
        config: typing.Mapping[str, typing.Any],
        parent: typing.Optional[Exception] = None,
    ):

        self._module_cls = module_cls
        self._config = config

        self._parent: typing.Optional[Exception] = parent

        if not msg.endswith("."):
            _msg = msg + "."
        else:
            _msg = msg

        super().__init__(_msg)


class ValueTypeConfigException(Exception):
    def __init__(
        self,
        msg: str,
        type_cls: typing.Type["ValueType"],
        config: typing.Mapping[str, typing.Any],
        parent: typing.Optional[Exception] = None,
    ):

        self._type_cls = type_cls
        self._config = config

        self._parent: typing.Optional[Exception] = parent

        if not msg.endswith("."):
            _msg = msg + "."
        else:
            _msg = msg

        super().__init__(_msg)


class KiaraValueException(Exception):
    def __init__(
        self,
        value_type: typing.Type["ValueType"],
        value_data: typing.Any,
        exception: Exception,
    ):
        self._value_type: typing.Type["ValueType"] = value_type
        self._value_data: typing.Any = value_data
        self._exception: Exception = exception

        exc_msg = str(self._exception)
        if not exc_msg:
            exc_msg = "no details available"

        super().__init__(f"Invalid value of type '{value_type._value_type_name}': {exc_msg}")  # type: ignore


class KiaraProcessingException(Exception):
    def __init__(
        self,
        msg: typing.Union[str, Exception],
        module: typing.Optional["KiaraModule"] = None,
        inputs: typing.Optional["ValueSet"] = None,
    ):
        self._module: typing.Optional["KiaraModule"] = module
        self._inputs: typing.Optional["ValueSet"] = inputs
        if isinstance(msg, Exception):
            self._parent: typing.Optional[Exception] = msg
            _msg = str(msg)
        else:
            self._parent = None
            _msg = msg
        super().__init__(_msg)

    @property
    def module(self) -> "KiaraModule":
        return self._module  # type: ignore

    @property
    def inputs(self) -> "ValueSet":
        return self._inputs  # type: ignore

    @property
    def parent_exception(self) -> typing.Optional[Exception]:
        return self._parent



# -*- coding: utf-8 -*-
import pytest

from kiara import Kiara
from kiara.data.values import Value, ValueSchema
from kiara.data.values.value_set import SlottedValueSet
from kiara.defaults import SpecialValue


def test_values_create(kiara: Kiara):

    value_schema = ValueSchema(type="string")
    v = kiara.data_registry.register_data(value_schema=value_schema)
    assert not v.is_set


def test_registry_values(kiara: Kiara):

    value_schema_1 = ValueSchema(type="string", required=False)

    reg = kiara.data_registry

    v = reg.register_data(value_schema=value_schema_1)
    with pytest.raises(Exception):
        v.get_value_data()

    v = reg.register_data(value_data=None, value_schema=value_schema_1)
    v.get_value_data()

    v = reg.register_data(value_data=SpecialValue.NO_VALUE, value_schema=value_schema_1)
    assert v.get_value_data() is None

    v = reg.register_data(value_data=SpecialValue.NOT_SET, value_schema=value_schema_1)
    with pytest.raises(Exception) as e:
        v.get_value_data()

    assert "Value not set" in str(e.value)

    v = reg.register_data(value_data="xxx", value_schema=value_schema_1)
    assert v.get_value_data() == "xxx"


def test_value_set_read_only(kiara: Kiara):

    schema_a = ValueSchema(type="boolean")
    schema_b = ValueSchema(type="boolean")

    value_set = SlottedValueSet.from_schemas(
        kiara=kiara, schemas={"a": schema_a, "b": schema_b}, title="TestOutput"
    )
    assert value_set.get_value_data("a") is None
    assert value_set.get_value_data("b") is None

    assert not value_set.items_are_valid()

    val_a = value_set.get_value_obj("a")
    assert val_a.get_value_data() is None

    with pytest.raises(Exception):
        value_set.set_value("a", False)


def test_value_set_impl_read_write(kiara: Kiara):

    schema_a = ValueSchema(type="boolean")
    schema_b = ValueSchema(type="boolean")

    value_set = SlottedValueSet.from_schemas(
        kiara=kiara,
        schemas={"a": schema_a, "b": schema_b},
        title="TestOutput",
        read_only=False,
    )
    assert value_set.get_value_data("a") is None
    assert value_set.get_value_data("b") is None

    assert not value_set.items_are_valid()

    val_a = value_set.get_value_obj("a")
    assert val_a.get_value_data() is None

    value_set.set_value("a", False)
    val_a_new = value_set.get_value_obj("a")
    assert val_a_new != val_a
    assert val_a_new.get_value_data() is False

    assert not value_set.items_are_valid()

    result = value_set.set_values(a=True, b=True)
    val_a_new_2 = result["a"]
    assert isinstance(val_a_new_2, Value)
    assert val_a_new != val_a_new_2
    assert val_a_new_2 != val_a

    assert value_set.items_are_valid()

    assert value_set.get_all_value_data() == {"a": True, "b": True}



# -*- coding: utf-8 -*-
import typing
from pydantic import BaseModel

from kiara.data import Value
from kiara.metadata.core_models import PythonClassMetadata
from kiara.operations.extract_metadata import ExtractMetadataModule


class ExtractPythonClass(ExtractMetadataModule):
    """Extract metadata about the Python type of a value."""

    _module_type_name = "metadata.python_class"

    @classmethod
    def _get_supported_types(cls) -> typing.Union[str, typing.Iterable[str]]:
        return "*"

    @classmethod
    def get_metadata_key(cls) -> str:
        return "python_class"

    def _get_metadata_schema(
        self, type: str
    ) -> typing.Union[str, typing.Type[BaseModel]]:

        return PythonClassMetadata

    def extract_metadata(self, value: Value) -> typing.Mapping[str, typing.Any]:

        item = value.get_value_data()
        cls = item.__class__

        return {
            "class_name": cls.__name__,
            "module_name": cls.__module__,
            "full_name": f"{cls.__module__}.{cls.__name__}",
        }



# -*- coding: utf-8 -*-
import typing
from pyarrow import Table
from pydantic import BaseModel, Field, root_validator
from rich import box
from rich.console import RenderableType
from rich.table import Table as RichTable

from kiara.interfaces import get_console
from kiara.utils import dict_from_cli_args


class OutputDetails(BaseModel):
    @classmethod
    def from_data(cls, data: typing.Any):

        if isinstance(data, str):
            if "=" in data:
                data = [data]
            else:
                data = [f"format={data}"]

        if isinstance(data, typing.Iterable):
            data = list(data)
            if len(data) == 1 and isinstance(data[0], str) and "=" not in data[0]:
                data = [f"format={data[0]}"]
            output_details_dict = dict_from_cli_args(*data)
        else:
            raise TypeError(
                f"Can't parse output detail config: invalid input type '{type(data)}'."
            )

        output_details = OutputDetails(**output_details_dict)
        return output_details

    format: str = Field(description="The output format.")
    target: str = Field(description="The output target.")
    config: typing.Dict[str, typing.Any] = Field(
        description="Output configuration.", default_factory=dict
    )

    @root_validator(pre=True)
    def _set_defaults(cls, values):

        target: str = values.pop("target", "terminal")
        format: str = values.pop("format", None)
        if format is None:
            if target == "terminal":
                format = "terminal"
            else:
                if target == "file":
                    format = "json"
                else:
                    ext = target.split(".")[-1]
                    if ext in ["yaml", "json"]:
                        format = ext
                    else:
                        format = "json"
        result = {"format": format, "target": target, "config": dict(values)}

        return result


def pretty_print_arrow_table(
    table: Table,
    rows_head: typing.Optional[int] = None,
    rows_tail: typing.Optional[int] = None,
    max_row_height: typing.Optional[int] = None,
    max_cell_length: typing.Optional[int] = None,
) -> RenderableType:

    rich_table = RichTable(box=box.SIMPLE)
    for cn in table.column_names:
        rich_table.add_column(cn)

    num_split_rows = 1

    if rows_head is not None:

        if rows_head < 0:
            rows_head = 0

        if rows_head > table.num_rows:
            rows_head = table.num_rows
            rows_tail = None
            num_split_rows = 0

        if rows_tail is not None:
            if rows_head + rows_tail >= table.num_rows:  # type: ignore
                rows_head = table.num_rows
                rows_tail = None
                num_split_rows = 0
    else:
        num_split_rows = 0

    if rows_head is not None:
        head = table.slice(0, rows_head)
        num_rows = rows_head
    else:
        head = table
        num_rows = table.num_rows

    table_dict = head.to_pydict()
    for i in range(0, num_rows):
        row = []
        for cn in table.column_names:
            cell = table_dict[cn][i]
            cell_str = str(cell)
            if max_row_height and max_row_height > 0 and "\n" in cell_str:
                lines = cell_str.split("\n")
                if len(lines) > max_row_height:
                    if max_row_height == 1:
                        lines = lines[0:1]
                    else:
                        half = int(len(lines) / 2)
                        lines = lines[0:half] + [".."] + lines[-half:]
                cell_str = "\n".join(lines)

            if max_cell_length and max_cell_length > 0:
                lines = []
                for line in cell_str.split("\n"):
                    if len(line) > max_cell_length:
                        line = line[0:(max_cell_length)] + " ..."
                    else:
                        line = line
                    lines.append(line)
                cell_str = "\n".join(lines)

            row.append(cell_str)

        rich_table.add_row(*row)

    if num_split_rows:
        for i in range(0, num_split_rows):
            row = []
            for _ in table.column_names:
                row.append("...")
            rich_table.add_row(*row)

    if rows_head:
        if rows_tail is not None:
            if rows_tail < 0:
                rows_tail = 0

            tail = table.slice(table.num_rows - rows_tail)
            table_dict = tail.to_pydict()
            for i in range(0, num_rows):

                row = []
                for cn in table.column_names:
                    cell = table_dict[cn][i]
                    cell_str = str(cell)

                    if max_row_height and max_row_height > 0 and "\n" in cell_str:
                        lines = cell_str.split("\n")
                        if len(lines) > max_row_height:
                            if max_row_height == 1:
                                lines = lines[0:1]
                            else:
                                half = int(len(lines) / 2)
                                lines = lines[0:half] + [".."] + lines[-half:]
                        cell_str = "\n".join(lines)

                    if max_cell_length and max_cell_length > 0:
                        lines = []
                        for line in cell_str.split("\n"):

                            if len(line) > max_cell_length:
                                line = line[0:(max_cell_length)] + " ..."
                            else:
                                line = line
                            lines.append(line)
                        cell_str = "\n".join(lines)

                    row.append(cell_str)

                rich_table.add_row(*row)

    return rich_table


def rich_print(msg: typing.Any = None) -> None:

    if msg is None:
        msg = ""
    console = get_console()
    console.print(msg)


def first_line(text: str):

    if "\n" in text:
        return text.split("\n")[0].strip()
    else:
        return text


def create_table_from_base_model(model_cls: typing.Type[BaseModel]):

    table = RichTable(box=box.SIMPLE)
    table.add_column("Field")
    table.add_column("Type")
    table.add_column("Description")
    table.add_column("Required")

    props = model_cls.schema().get("properties", {})

    for field_name, field in model_cls.__fields__.items():
        row = [field_name]
        p = props.get(field_name, None)
        p_type = None
        if p is not None:
            p_type = p.get("type", None)
            # TODO: check 'anyOf' keys

        if p_type is None:
            p_type = "-- check source --"
        row.append(p_type)
        desc = p.get("description", "")
        row.append(desc)
        row.append("yes" if field.required else "no")
        table.add_row(*row)

    return table



# -*- coding: utf-8 -*-
import logging
import sys
import typing
import uuid
from rich.console import Console, ConsoleOptions, RenderResult

from kiara.data.registry import DataRegistry
from kiara.data.values import ValueSlot
from kiara.data.values.value_set import SlottedValueSet, ValueSet
from kiara.defaults import SpecialValue
from kiara.events import (
    PipelineInputEvent,
    PipelineOutputEvent,
    StepEvent,
    StepInputEvent,
    StepOutputEvent,
)
from kiara.pipeline import PipelineValuesInfo, StepStatus
from kiara.pipeline.controller import PipelineController
from kiara.pipeline.controller.batch import BatchController
from kiara.pipeline.listeners import PipelineListener
from kiara.pipeline.structure import PipelineStep, PipelineStructure
from kiara.pipeline.values import (
    PipelineInputRef,
    PipelineOutputRef,
    StepInputRef,
    StepOutputRef,
    ValueRef,
)

if typing.TYPE_CHECKING:
    from kiara.info.pipelines import PipelineState
    from kiara.kiara import Kiara

log = logging.getLogger("kiara")


class Pipeline(object):
    """An instance of a [PipelineStructure][kiara.pipeline.structure.PipelineStructure] that holds state for all of the inputs/outputs of the steps within."""

    def __init__(
        self,
        structure: PipelineStructure,
        # constants: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        controller: typing.Optional[PipelineController] = None,
        title: typing.Optional[str] = None,
    ):

        self._id: str = str(uuid.uuid4())
        if title is None:
            title = self._id

        self._title: str = title
        self._structure: PipelineStructure = structure

        self._pipeline_inputs: ValueSet = None  # type: ignore
        self._pipeline_outputs: ValueSet = None  # type: ignore

        self._step_inputs: typing.Mapping[str, ValueSet] = None  # type: ignore
        self._step_outputs: typing.Mapping[str, ValueSet] = None  # type: ignore

        self._value_refs: typing.Mapping[ValueSlot, typing.Iterable[ValueRef]] = None  # type: ignore
        self._status: StepStatus = StepStatus.STALE
        self._steps_by_stage: typing.Optional[
            typing.Dict[int, typing.Dict[str, PipelineStep]]
        ] = None
        self._inputs_by_stage: typing.Optional[
            typing.Dict[int, typing.List[str]]
        ] = None

        self._kiara: "Kiara" = self._structure._kiara
        self._data_registry: DataRegistry = self._kiara.data_registry

        self._init_values()

        if controller is None:
            controller = BatchController(self, kiara=self._kiara)
        else:
            controller.set_pipeline(self)
        self._controller: PipelineController = controller

        self._listeners: typing.List[PipelineListener] = []

        self._update_status()

    def __eq__(self, other):

        if not isinstance(other, Pipeline):
            return False

        return self._id == other._id

    def __hash__(self):

        return hash(self._id)

    @property
    def id(self) -> str:
        return self._id

    @property
    def title(self) -> str:
        return self._title

    @property
    def structure(self) -> PipelineStructure:
        return self._structure

    @property
    def controller(self) -> PipelineController:
        if self._controller is None:
            raise Exception("No controller set (yet).")
        return self._controller

    @property
    def inputs(self) -> ValueSet:
        return self._pipeline_inputs

    @property
    def outputs(self) -> ValueSet:
        return self._pipeline_outputs

    # def set_pipeline_inputs(self, **inputs: typing.Any):
    #     self._controller.set_pipeline_inputs(**inputs)

    @property
    def step_ids(self) -> typing.Iterable[str]:
        return self._structure.step_ids

    def get_step(self, step_id: str) -> PipelineStep:
        return self._structure.get_step(step_id)

    def get_steps_by_stage(
        self,
    ) -> typing.Mapping[int, typing.Mapping[str, PipelineStep]]:

        if self._steps_by_stage is not None:
            return self._steps_by_stage

        result: typing.Dict[int, typing.Dict[str, PipelineStep]] = {}
        for step_id in self.step_ids:
            step = self.get_step(step_id)
            stage = step.processing_stage
            assert stage is not None
            result.setdefault(stage, {})[step_id] = step

        self._steps_by_stage = result
        return self._steps_by_stage

    def get_inputs_by_stage(self) -> typing.Mapping[int, typing.Iterable[str]]:

        if self._inputs_by_stage is not None:
            return self._inputs_by_stage

        result: typing.Dict[int, typing.List[str]] = {}
        for k, v in self.inputs._value_slots.items():  # type: ignore
            refs = self._value_refs[v]
            min_stage = sys.maxsize
            for ref in refs:
                if not isinstance(ref, StepInputRef):
                    continue
                step = self.get_step(ref.step_id)
                stage = step.processing_stage
                assert stage is not None
                if stage < min_stage:
                    min_stage = stage  # type: ignore
                result.setdefault(min_stage, []).append(k)

        self._inputs_by_stage = result
        return self._inputs_by_stage

    def get_step_inputs(self, step_id: str) -> ValueSet:
        return self._step_inputs[step_id]

    def get_step_outputs(self, step_id: str) -> ValueSet:
        return self._step_outputs[step_id]

    def add_listener(self, listener: PipelineListener) -> None:
        self._listeners.append(listener)

    @property
    def status(self) -> StepStatus:
        return self._state

    def _update_status(self):

        if self.inputs is None:
            new_state = StepStatus.STALE
        elif not self.inputs.items_are_valid():
            new_state = StepStatus.STALE
        elif not self.outputs.items_are_valid():
            new_state = StepStatus.INPUTS_READY
        else:
            new_state = StepStatus.RESULTS_READY

        self._state = new_state

    def _init_values(self):
        """Initialize this object. This should only be called once.

        Basically, this goes through all the inputs and outputs of all steps, and 'allocates' a PipelineValueInfo object
        for each of them. In case where output/input or pipeline-input/input points are connected, only one
        value item is allocated, since those refer to the same value.
        """

        pipeline_inputs: typing.Dict[str, ValueSlot] = {}
        pipeline_outputs: typing.Dict[str, ValueSlot] = {}

        all_step_inputs: typing.Dict[str, typing.Dict[str, ValueSlot]] = {}
        all_step_outputs: typing.Dict[str, typing.Dict[str, ValueSlot]] = {}

        value_refs: typing.Dict[ValueSlot, typing.List[ValueRef]] = {}

        # create the value objects that are associated with step outputs
        # all pipeline outputs are created here too, since the only place
        # those can be associated are step outputs
        for step_id, step_details in self._structure.steps_details.items():

            step_outputs: typing.Mapping[str, StepOutputRef] = step_details["outputs"]

            for output_name, output_point in step_outputs.items():

                init_output_value_item = self._data_registry.register_data(
                    value_schema=output_point.value_schema
                )
                output_value_slot = self._data_registry.register_alias(
                    value_or_schema=init_output_value_item, callbacks=[self]
                )
                value_refs.setdefault(output_value_slot, []).append(output_point)

                all_step_outputs.setdefault(step_id, {})[
                    output_name
                ] = output_value_slot

                # not all step outputs necessarily need to be connected to a pipeline output
                if output_point.pipeline_output:

                    pipeline_outputs[output_point.pipeline_output] = output_value_slot
                    po = self._structure.pipeline_outputs[output_point.pipeline_output]
                    value_refs.setdefault(output_value_slot, []).append(po)

        # create the value objects that are associated with step inputs
        for step_id, step_details in self._structure.steps_details.items():

            step_inputs: typing.Mapping[str, StepInputRef] = step_details["inputs"]

            for input_name, input_point in step_inputs.items():

                # if this step input gets fed from a pipeline_input (meaning user input in most cases),
                # we need to create a DataValue for that pipeline input
                # vm = ValueMetadata(
                #     origin=f"{self.id}.steps.{step_id}.inputs.{input_point.value_name}"
                # )
                if input_point.connected_pipeline_input:
                    connected_pipeline_input_name = input_point.connected_pipeline_input
                    pipeline_input_field: PipelineInputRef = (
                        self._structure.pipeline_inputs[connected_pipeline_input_name]
                    )
                    pipeline_input_slot: ValueSlot = pipeline_inputs.get(
                        connected_pipeline_input_name, None
                    )

                    if pipeline_input_slot is None:
                        # if the pipeline input wasn't created by another step input before,
                        # we need to take care of it here

                        if pipeline_input_field.is_constant:
                            init_value = self.structure.constants[
                                pipeline_input_field.value_name
                            ]
                        else:
                            init_value = self.structure.defaults.get(
                                pipeline_input_field.value_name, SpecialValue.NOT_SET
                            )

                        init_pipeline_input_value = self._data_registry.register_data(
                            value_data=init_value,
                            value_schema=pipeline_input_field.value_schema,
                        )
                        # TODO: check whether it's a constant?
                        pipeline_input_slot = self._data_registry.register_alias(
                            value_or_schema=init_pipeline_input_value, callbacks=[self]
                        )
                        value_refs.setdefault(pipeline_input_slot, []).append(
                            pipeline_input_field
                        )

                        pipeline_inputs[
                            connected_pipeline_input_name
                        ] = pipeline_input_slot

                    all_step_inputs.setdefault(step_id, {})[
                        input_name
                    ] = pipeline_input_slot
                    value_refs.setdefault(pipeline_input_slot, []).append(input_point)

                elif input_point.connected_outputs:

                    for co in input_point.connected_outputs:
                        if len(input_point.connected_outputs) == 1 and not co.sub_value:
                            # this means the input is the same value as the connected output
                            output_value: ValueSlot = all_step_outputs[co.step_id][
                                co.value_name
                            ]
                            all_step_inputs.setdefault(input_point.step_id, {})[
                                input_point.value_name
                            ] = output_value
                            value_refs.setdefault(output_value, []).append(input_point)
                        else:
                            print(input_point.connected_outputs)
                            raise NotImplementedError()
                            # sub_value = co.sub_value

                            # linked_values = {}
                            # for co in input_point.connected_outputs:
                            #     output_value = all_step_outputs[co.step_id][co.value_name]
                            #     sub_value = co.sub_value
                            #     if len(input_point.connected_outputs) > 1 and not sub_value:
                            #         raise NotImplementedError()
                            #         sub_value = {"config": co.step_id}
                            #     if sub_value is not None:
                            #         raise NotImplementedError
                            #
                            #     linked_values[output_value.id] = sub_value
                            #
                            # step_input = self._data_registry.register_linked_value(
                            #     parent_id=self.id,
                            #     linked_values=linked_values,
                            #     value_schema=input_point.value_schema,
                            #     value_refs=input_point,
                            # )
                            # self._data_registry.register_callback(
                            #     self.values_updated, step_input
                            # )
                            # all_step_inputs.setdefault(input_point.step_id, {})[
                            #     input_point.value_name
                            # ] = step_input

                else:
                    raise Exception(
                        f"Invalid value point type for this location: {input_point}"
                    )

        if not pipeline_inputs:
            raise Exception(f"Can't init pipeline '{self.title}': no pipeline inputs")

        self._pipeline_inputs = SlottedValueSet(
            items=pipeline_inputs,
            read_only=False,
            title=f"Inputs for pipeline '{self.title}'",
            kiara=self._kiara,
            registry=self._data_registry,
        )
        if not pipeline_outputs:
            raise Exception(f"Can't init pipeline '{self.title}': no pipeline outputs")

        self._pipeline_outputs = SlottedValueSet(
            items=pipeline_outputs,
            read_only=True,
            title=f"Outputs for pipeline '{self.title}'",
            kiara=self._kiara,
            registry=self._data_registry,
        )
        self._step_inputs = {}
        for step_id, inputs in all_step_inputs.items():
            self._step_inputs[step_id] = SlottedValueSet(
                items=inputs,
                read_only=True,
                title=f"Inputs for step '{step_id}' of pipeline '{self.title}",
                kiara=self._kiara,
                registry=self._data_registry,
            )
        self._step_outputs = {}
        for step_id, outputs in all_step_outputs.items():
            self._step_outputs[step_id] = SlottedValueSet(
                read_only=False,
                items=outputs,
                title=f"Outputs for step '{step_id}' of pipeline '{self.title}'",
                kiara=self._kiara,
                registry=self._data_registry,
            )

        self._value_refs = value_refs
        self._steps_by_stage = None
        self._inputs_by_stage = None

    def values_updated(self, *items: ValueSlot) -> None:

        updated_inputs: typing.Dict[str, typing.List[str]] = {}
        updated_outputs: typing.Dict[str, typing.List[str]] = {}
        updated_pipeline_inputs: typing.List[str] = []
        updated_pipeline_outputs: typing.List[str] = []

        # print("===================================================")
        # for item in items:
        #     print(item)
        # print("===================================================")

        self._update_status()

        if self._value_refs is None:
            # means init is not finished yet
            return

        for item in items:

            # TODO: multiple value fields, also check pipeline id
            references = self._value_refs.get(item, None)
            assert references

            for p in references:

                if isinstance(p, StepInputRef):
                    updated_inputs.setdefault(p.step_id, []).append(p.value_name)
                elif isinstance(p, StepOutputRef):
                    updated_outputs.setdefault(p.step_id, []).append(p.value_name)
                elif isinstance(p, PipelineInputRef):
                    updated_pipeline_inputs.append(p.value_name)
                elif isinstance(p, PipelineOutputRef):
                    updated_pipeline_outputs.append(p.value_name)
                else:
                    raise TypeError(f"Can't update, invalid type: {type(p)}")

        # print('========================================')
        # print('---')
        # print("Upaded pipeline input")
        # print(updated_pipeline_inputs)
        # print('---')
        # print("Upaded step inputs")
        # print(updated_inputs)
        # print('---')
        # print("Upaded step outputs")
        # print(updated_outputs)
        # print('---')
        # print("Upaded pipeline outputs")
        # print(updated_pipeline_outputs)

        if updated_pipeline_inputs:
            event_pi = PipelineInputEvent(
                pipeline_id=self.id,
                updated_pipeline_inputs=updated_pipeline_inputs,
            )
            self._controller.pipeline_inputs_changed(event_pi)
            self._notify_pipeline_listeners(event_pi)

        if updated_outputs:
            event_so = StepOutputEvent(
                pipeline_id=self.id,
                updated_step_outputs=updated_outputs,
            )
            self._controller.step_outputs_changed(event_so)
            self._notify_pipeline_listeners(event_so)

        if updated_inputs:
            event_si = StepInputEvent(
                pipeline_id=self.id,
                updated_step_inputs=updated_inputs,
            )
            self._controller.step_inputs_changed(event_si)
            self._notify_pipeline_listeners(event_si)

        if updated_pipeline_outputs:
            event_po = PipelineOutputEvent(
                pipeline_id=self.id,
                updated_pipeline_outputs=updated_pipeline_outputs,
            )
            self._controller.pipeline_outputs_changed(event_po)
            self._notify_pipeline_listeners(event_po)

    def _notify_pipeline_listeners(self, event: StepEvent):

        for listener in self._listeners:
            if event.type == "step_input":  # type: ignore
                listener.step_inputs_changed(event)  # type: ignore
            elif event.type == "step_output":  # type: ignore
                listener.step_outputs_changed(event)  # type: ignore
            elif event.type == "pipeline_input":  # type: ignore
                listener.pipeline_inputs_changed(event)  # type: ignore
            elif event.type == "pipeline_output":  # type: ignore
                listener.pipeline_outputs_changed(event)  # type: ignore
            else:
                raise Exception(f"Unsupported type: {event.type}")  # type: ignore

    def get_current_state(self) -> "PipelineState":

        step_inputs = {}
        step_states = {}
        for k, v in self._step_inputs.items():
            step_inputs[k] = PipelineValuesInfo.from_value_set(v)
            if v.items_are_valid():
                step_states[k] = StepStatus.INPUTS_READY
            else:
                step_states[k] = StepStatus.STALE

        step_outputs = {}
        for k, v in self._step_outputs.items():
            step_outputs[k] = PipelineValuesInfo.from_value_set(v)
            if v.items_are_valid():
                step_states[k] = StepStatus.RESULTS_READY

        from kiara.info.pipelines import PipelineState

        state = PipelineState(
            structure=self.structure.to_details(),
            pipeline_inputs=self._pipeline_inputs.to_details(),
            pipeline_outputs=self._pipeline_outputs.to_details(),
            step_states=step_states,
            step_inputs=step_inputs,
            step_outputs=step_outputs,
            status=self.status,
        )
        return state

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        yield self.get_current_state()



# -*- coding: utf-8 -*-
import typing

try:
    from typing import Literal
except Exception:
    from typing_extensions import Literal  # type: ignore

from pydantic import BaseModel, Field


class StepEvent(BaseModel):
    class Config:
        allow_mutation = False

    pipeline_id: str

    def __repr__(self):
        d = self.dict()
        d.pop("pipeline_id")
        return f"{self.__class__.__name__}(pipeline_id={self.pipeline_id} data={d}"

    def __str__(self):
        return self.__repr__()


class StepInputEvent(StepEvent):
    """Event that gets fired when one or several inputs for steps within a pipeline have changed."""

    type: Literal["step_input"] = "step_input"
    updated_step_inputs: typing.Dict[str, typing.List[str]] = Field(
        description="steps (keys) with updated inputs which need re-processing (value is list of updated input names)"
    )

    @property
    def newly_stale_steps(self) -> typing.List[str]:
        """Convenience method to display the steps that have been rendered 'stale' by this event."""
        return list(self.updated_step_inputs.keys())


class StepOutputEvent(StepEvent):
    """Event that gets fired when one or several outputs for steps within a pipeline have changed."""

    type: Literal["step_output"] = "step_output"
    updated_step_outputs: typing.Dict[str, typing.List[str]] = Field(
        description="steps (keys) that finished processing of one, several or all outputs (values are list of 'finished' output fields)"
    )


class PipelineInputEvent(StepEvent):
    """Event that gets fired when one or several inputs for the pipeline itself have changed."""

    type: Literal["pipeline_input"] = "pipeline_input"
    updated_pipeline_inputs: typing.List[str] = Field(
        description="list of pipeline input names that where changed"
    )


class PipelineOutputEvent(StepEvent):
    """Event that gets fired when one or several outputs for the pipeline itself have changed."""

    type: Literal["pipeline_output"] = "pipeline_output"
    updated_pipeline_outputs: typing.List[str] = Field(
        description="list of pipeline output names that where changed"
    )


# class OtherEvent(StepEvent):
#
#     new_streaming_input: typing.Dict[str, typing.List[str]] = Field(
#         description="steps (keys) where there was new data streamed to one or more inputs (values are list of those input names)"
#     )



# -*- coding: utf-8 -*-
"""The 'run' subcommand for the cli."""

import click
import os.path
import sys
import typing
from pathlib import Path
from rich import box
from rich.console import Console, RenderGroup
from rich.panel import Panel
from rich.syntax import Syntax

from kiara import Kiara
from kiara.data.values.value_set import ValueSet, ValuesInfo
from kiara.defaults import DEFAULT_TO_JSON_CONFIG
from kiara.interfaces.cli.utils import _create_module_instance
from kiara.module import KiaraModule
from kiara.pipeline import StepStatus
from kiara.pipeline.controller.batch import BatchController
from kiara.processing import JobStatus
from kiara.utils import create_table_from_field_schemas, dict_from_cli_args, is_debug
from kiara.utils.output import OutputDetails, rich_print


@click.command()
@click.argument("module", nargs=1, metavar="MODULE_OR_OPERATION")
@click.argument("inputs", nargs=-1, required=False)
@click.option(
    "--module-config",
    "-c",
    required=False,
    help="(Optional) module configuration.",
    multiple=True,
)
@click.option(
    "--explain",
    "-e",
    help="Display additional workflow details.",
    is_flag=True,
)
@click.option(
    "--output", "-o", help="The output format and configuration.", multiple=True
)
@click.option(
    "--save",
    "-s",
    help="Save one or several of the outputs of this run. If the argument contains a '=', the format is [output_name]=[alias], if not, the values will be saved as '[alias]-[output_name]'.",
    required=False,
    multiple=True,
)
@click.pass_context
def run(ctx, module, inputs, module_config, output, explain, save):
    """Execute a workflow run."""

    if module_config:
        module_config = dict_from_cli_args(*module_config)
    else:
        module_config = {}

    if not save:
        aliases: typing.Dict[str, typing.List[str]] = {}
        full_aliases: typing.List[str] = []
    else:
        aliases = {}
        full_aliases = []
        for a in save:
            if "=" not in a:
                full_aliases.append(a)
            else:
                tokens = a.split("=")
                if len(tokens) != 2:
                    print()
                    print(f"Invalid alias format, can only contain a single '=': {a}")
                    sys.exit(1)

                aliases.setdefault(tokens[0], []).append(tokens[1])

    kiara_obj: Kiara = ctx.obj["kiara"]

    if module in kiara_obj.available_module_types:
        module_name = module
    elif module in kiara_obj.operation_mgmt.profiles.keys():
        op_config = kiara_obj.operation_mgmt.profiles[module]
        module_name = op_config.module_type
        module_config = op_config.module_config
    elif f"core.{module}" in kiara_obj.available_module_types:
        module_name = f"core.{module}"
    elif os.path.isfile(module):
        module_name = kiara_obj.register_pipeline_description(
            module, raise_exception=True
        )
    else:
        rich_print(
            f"\nInvalid module name '[i]{module}[/i]'. Must be a path to a pipeline file, or one of the available modules:\n"
        )
        for n in kiara_obj.available_module_types:
            rich_print(f"  - [i]{n}[/i]")
        sys.exit(1)

    if not inputs:

        module_obj: KiaraModule = _create_module_instance(
            ctx=ctx, module_type=module_name, module_config=module_config
        )

        one_required = False
        for input_name in module_obj.input_names:
            if module_obj.input_required(input_name):
                one_required = True
                break

        if one_required:

            inputs_table = create_table_from_field_schemas(
                _show_header=True, **module_obj.input_schemas
            )
            print()
            print(
                "No inputs provided, not running the workflow. To run it, provide input following this schema:"
            )
            rich_print(inputs_table)
            sys.exit(0)

    output_details = OutputDetails.from_data(output)
    silent = False
    if output_details.format == "silent":
        silent = True

    force_overwrite = output_details.config.get("force", False)

    # SUPPORTED_TARGETS = ["terminal", "file"]
    # if output_details.target not in SUPPORTED_TARGETS:
    #     print()
    #     rich_print(f"Invalid output target '{output_details.target}', must be one of: [i]{', '.join(SUPPORTED_TARGETS)}[/i]")
    #     sys.exit(1)

    target_file: typing.Optional[Path] = None
    if output_details.target != "terminal":
        if output_details.target == "file":
            target_dir = Path.cwd()
            target_file = target_dir / f"{module_name}.{output_details.format}"
        else:
            target_file = Path(
                os.path.realpath(os.path.expanduser(output_details.target))
            )

        if target_file.exists() and not force_overwrite:
            print()
            print(
                f"Can't run workflow, the target files already exist, and '--output force=true' not specified: {target_file}"
            )
            sys.exit(1)

    # from kiara.processing.parallel import ThreadPoolProcessor
    # processor = ThreadPoolProcessor()
    processor = None
    controller = BatchController(
        processor=processor, auto_process=False, kiara=kiara_obj
    )

    # TODO: should we let the user specify?
    workflow_id = None
    if workflow_id is None:
        workflow_id = f"{module_name}_0"

    workflow = kiara_obj.create_workflow(
        module_name,
        module_config=module_config,
        workflow_id=workflow_id,
        controller=controller,
    )

    if save:
        final_aliases = {}
        invalid_fields = []
        for field_name, alias in aliases.items():
            if field_name not in workflow.outputs.get_all_field_names():
                invalid_fields.append(field_name)
            else:
                final_aliases[field_name] = alias

        for alias in full_aliases:
            for field_name in workflow.outputs.get_all_field_names():
                final_aliases.setdefault(field_name, []).append(f"{alias}-{field_name}")

        if invalid_fields:
            print()
            print(
                f"Can't run workflow, invalid field name(s) when specifying aliases: {', '.join(invalid_fields)}. Valid field names: {', '.join(workflow.outputs.get_all_field_names())}"
            )
            sys.exit(1)

    list_keys = []

    for name, value in workflow.inputs.items():
        if value.value_schema.type in ["array", "list"]:
            list_keys.append(name)

    workflow_input = dict_from_cli_args(*inputs, list_keys=list_keys)

    try:
        if workflow_input:
            workflow.inputs.set_values(**workflow_input)

    except Exception as e:
        import traceback

        traceback.print_exc()
        print()
        rich_print(f"[bold red]Input value error[/bold red]: {e}")
        sys.exit(1)

    try:
        workflow.controller.process_pipeline()
    except Exception as e:
        print()
        print(e)

    if explain:
        print()
        kiara_obj.explain(workflow.current_state)

        if workflow.status == StepStatus.RESULTS_READY:
            vi = ValuesInfo(workflow.outputs)
            vi_table = vi.create_value_info_table(
                ensure_metadata=True, show_headers=True
            )
            panel = Panel(Panel(vi_table), box=box.SIMPLE)
            rich_print("[b]Output data details[/b]")
            rich_print(panel)

    if workflow.status != StepStatus.RESULTS_READY:

        failed = {}
        for step_id in workflow.pipeline.step_ids:
            job = workflow.controller.get_job_details(step_id)
            if not job:
                continue
            if job.status == JobStatus.FAILED:
                failed[step_id] = job.error if job.error else "-- no error details --"

        print()
        if failed:
            rich_print(
                "[bold red]Error:[/bold red] One or several workflow steps failed!\n"
            )
            for s_id, msg in failed.items():
                rich_print(f" - [bold]{s_id}[/bold]: {msg}")

            sys.exit(3)
        else:
            rich_print(
                "Workflow results not ready: one or several inputs missing or invalid."
            )
            sys.exit(3)
    else:
        if not silent:

            if output_details.target == "terminal":
                if output_details.format == "terminal":
                    print()
                    all_renderables = []

                    for field_name, value in workflow.outputs.items():
                        try:
                            renderables = kiara_obj.pretty_print(value, "renderables")
                        except Exception as e:
                            if is_debug():
                                print(e)
                            renderables = [str(value.get_value_data())]

                        if isinstance(renderables, str) or not isinstance(
                            renderables, typing.Iterable
                        ):
                            renderables = [renderables]
                        p = Panel(
                            RenderGroup(*renderables),
                            box=box.ROUNDED,
                            title=f"Value: [b i]{field_name}[/b i]",
                            title_align="left",
                        )
                        all_renderables.append(p)

                    output = Panel(RenderGroup(*all_renderables), box=box.SIMPLE)
                    rich_print("[b]Output data[/b]")
                    rich_print(output)

                else:
                    raise NotImplementedError()

                    format = output_details.format
                    available_formats = kiara_obj.get_convert_target_types(
                        source_type="value_set"
                    )
                    if format not in available_formats:
                        print()
                        print(
                            f"Can't convert to output format '{format}', this format is not supported. Available formats: {', '.join(available_formats)}."
                        )
                        sys.exit(1)

                    config = {}
                    config.update(DEFAULT_TO_JSON_CONFIG)

                    try:
                        transformed: ValueSet = kiara_obj.transform_data(
                            workflow.outputs,
                            source_type="value_set",
                            target_type=format,
                            config=config,
                        )
                        transformed_value = transformed.get_value_data("target_value")

                        if format in ["json", "yaml"]:
                            transformed_str = Syntax(
                                transformed_value,
                                lexer_name=format,
                                background_color="default",
                            )
                            rich_print(transformed_str)
                        else:
                            print(transformed_value)
                    except Exception as e:
                        print()
                        rich_print(f"Can't transform outputs into '{format}': {e}")
                        sys.exit(1)

            else:
                raise NotImplementedError()

                if output_details.format == "terminal":

                    renderables = kiara_obj.pretty_print(value=value)
                    output = Panel(RenderGroup(*renderables), box=box.SIMPLE)
                    with open(target_file, "wt") as f:
                        console = Console(record=True, file=f)
                        console.print(output)
                else:

                    format = output_details.format
                    available_formats = kiara_obj.get_convert_target_types(
                        source_type="value_set"
                    )
                    if format not in available_formats:
                        print()
                        print(
                            f"Can't convert to output format '{format}', this format is not supported. Available formats: {', '.join(available_formats)}."
                        )
                        sys.exit(1)

                    config = {}
                    config.update(DEFAULT_TO_JSON_CONFIG)

                    transformed = kiara_obj.transform_data(
                        workflow.outputs,
                        source_type="value_set",
                        target_type=format,
                        config=config,
                    )
                    transformed_value = transformed.get_value_data()

                    target_file.parent.mkdir(parents=True, exist_ok=True)
                    # TODO: check whether to write text or bytes
                    target_file.write_text(transformed_value)

        if save:

            for field_name, aliases in final_aliases.items():
                rich_print(f"Saving '[i]{field_name}[/i]'...")
                try:
                    value = workflow.outputs.get_value_obj(field_name)
                    value_md = value.save(aliases=aliases)
                    msg = f"   -> done, id: [i]{value_md.id}[/i]"
                    if aliases:
                        msg = msg + f", aliases: [i]{', '.join(aliases)}[/i]"
                    rich_print(msg)
                except Exception as e:
                    if is_debug():
                        import traceback

                        traceback.print_exc()
                    rich_print(f"   -> failed: [red]{e}[/red]")
                print()



# -*- coding: utf-8 -*-
import pytest

from kiara import Kiara


def test_pipeline_default_controller(kiara: Kiara):

    pipeline = kiara.create_pipeline("logic.nand")

    pipeline.inputs.set_values(logic_nand__a=True, logic_nand__b=True)

    result = pipeline.outputs.get_all_value_data()
    assert result == {"logic_nand__y": False}


def test_pipeline_default_controller_invalid_inputs(kiara: Kiara):

    pipeline = kiara.create_pipeline("logic.nand")

    with pytest.raises(Exception) as e:
        pipeline.inputs.set_values(logic_nand__a1=True, logic_nand__b=True)

    assert "logic_nand__a1" in str(e.value)
    assert "logic_nand__b" in str(e.value)


def test_pipeline_default_controller_invalid_outputs(kiara: Kiara):

    pipeline = kiara.create_pipeline("logic.nand")

    pipeline.inputs.set_values(logic_nand__a=True, logic_nand__b=True)

    with pytest.raises(KeyError) as e:
        pipeline.outputs.get_value_data("logic_nand__y1")

    assert "logic_nand__y1" in str(e.value)
    assert "logic_nand__y" in str(e.value)



# -*- coding: utf-8 -*-
import logging
import typing

from kiara.pipeline.controller import PipelineController
from kiara.utils import is_debug

if typing.TYPE_CHECKING:
    from kiara.events import PipelineInputEvent, PipelineOutputEvent, StepInputEvent
    from kiara.kiara import Kiara
    from kiara.pipeline.pipeline import Pipeline
    from kiara.processing.processor import ModuleProcessor

log = logging.getLogger("kiara")


class BatchController(PipelineController):
    """A [PipelineController][kiara.pipeline.controller.PipelineController] that executes all pipeline steps non-interactively.

    This is the default implementation of a ``PipelineController``, and probably the most simple implementation of one.
    It waits until all inputs are set, after which it executes all pipeline steps in the required order.

    Arguments:
        pipeline: the pipeline to control
        auto_process: whether to automatically start processing the pipeline as soon as the input set is valid
    """

    def __init__(
        self,
        pipeline: typing.Optional["Pipeline"] = None,
        auto_process: bool = True,
        processor: typing.Optional["ModuleProcessor"] = None,
        kiara: typing.Optional["Kiara"] = None,
    ):

        self._auto_process: bool = auto_process
        self._is_running: bool = False
        super().__init__(pipeline=pipeline, processor=processor, kiara=kiara)

    @property
    def auto_process(self) -> bool:
        return self._auto_process

    @auto_process.setter
    def auto_process(self, auto_process: bool):
        self._auto_process = auto_process

    def process_pipeline(self):

        if self._is_running:
            log.debug("Pipeline running, doing nothing.")
            raise Exception("Pipeline already running.")

        self._is_running = True
        try:
            for stage in self.processing_stages:
                job_ids = []
                for step_id in stage:
                    if not self.can_be_processed(step_id):
                        if self.can_be_skipped(step_id):
                            continue
                        else:
                            raise Exception(
                                f"Required pipeline step '{step_id}' can't be processed, inputs not ready yet: {', '.join(self.invalid_inputs(step_id))}"
                            )
                    try:
                        job_id = self.process_step(step_id)
                        job_ids.append(job_id)
                    except Exception as e:
                        # TODO: cancel running jobs?
                        if is_debug():
                            import traceback

                            traceback.print_exc()
                        log.error(
                            f"Processing of step '{step_id}' from pipeline '{self.pipeline.id}' failed: {e}"
                        )
                        return False

                self._processor.wait_for(*job_ids)
                # for j_id in job_ids:
                #     job = self._processor.get_job_details(j_id)
                #     assert job is not None
                #     if job.error:
                #         print(job.error)
        finally:
            self._is_running = False

    def step_inputs_changed(self, event: "StepInputEvent"):

        if self._is_running:
            log.debug("Pipeline running, doing nothing.")
            return

        if not self.pipeline_is_ready():
            log.debug(f"Pipeline not ready after input event: {event}")
            return

        if self._auto_process:
            self.process_pipeline()

    def pipeline_outputs_changed(self, event: "PipelineOutputEvent"):

        if self.pipeline_is_finished():
            # TODO: check if something is running
            self._is_running = False


class BatchControllerManual(PipelineController):
    """A [PipelineController][kiara.pipeline.controller.PipelineController] that executes all pipeline steps non-interactively.

    This is the default implementation of a ``PipelineController``, and probably the most simple implementation of one.
    It waits until all inputs are set, after which it executes all pipeline steps in the required order.

    Arguments:
        pipeline: the pipeline to control
        auto_process: whether to automatically start processing the pipeline as soon as the input set is valid
    """

    def __init__(
        self,
        pipeline: typing.Optional["Pipeline"] = None,
        processor: typing.Optional["ModuleProcessor"] = None,
        kiara: typing.Optional["Kiara"] = None,
    ):

        self._finished_until: typing.Optional[int] = None
        self._is_running: bool = False
        super().__init__(pipeline=pipeline, processor=processor, kiara=kiara)

    @property
    def last_finished_stage(self) -> int:

        if self._finished_until is None:
            return 0
        else:
            return self._finished_until

    def process_pipeline(self):

        if self._is_running:
            log.debug("Pipeline running, doing nothing.")
            raise Exception("Pipeline already running.")

        self._is_running = True
        try:
            for stage in self.processing_stages:
                job_ids = []
                for step_id in stage:
                    if not self.can_be_processed(step_id):
                        if self.can_be_skipped(step_id):
                            continue
                        else:
                            raise Exception(
                                f"Required pipeline step '{step_id}' can't be processed, inputs not ready yet: {', '.join(self.invalid_inputs(step_id))}"
                            )
                    try:
                        job_id = self.process_step(step_id)
                        job_ids.append(job_id)
                    except Exception as e:
                        # TODO: cancel running jobs?
                        if is_debug():
                            import traceback

                            traceback.print_stack()
                        log.error(
                            f"Processing of step '{step_id}' from pipeline '{self.pipeline.title}' failed: {e}"
                        )
                        return False
                self._processor.wait_for(*job_ids)
        finally:
            self._is_running = False

    def step_inputs_changed(self, event: "StepInputEvent"):

        if self._is_running:
            log.debug("Pipeline running, doing nothing.")
            return

        if not self.pipeline_is_ready():
            log.debug(f"Pipeline not ready after input event: {event}")
            return

    def pipeline_inputs_changed(self, event: "PipelineInputEvent"):

        self._finished_until = None

    def pipeline_outputs_changed(self, event: "PipelineOutputEvent"):

        if self.pipeline_is_finished():
            # TODO: check if something is running
            self._is_running = False

    def process_stage(self, stage_nr: int):

        if self._is_running:
            log.debug("Pipeline running, doing nothing.")
            raise Exception("Pipeline already running.")

        self._is_running = True
        try:
            for idx, stage in enumerate(self.processing_stages):

                if idx + 1 > stage_nr:
                    break

                if self._finished_until is not None and idx <= self._finished_until:
                    continue

                job_ids = []
                for step_id in stage:
                    if not self.can_be_processed(step_id):
                        if self.can_be_skipped(step_id):
                            continue
                        else:
                            raise Exception(
                                f"Required pipeline step '{step_id}' can't be processed, inputs not ready yet: {', '.join(self.invalid_inputs(step_id))}"
                            )
                    try:
                        job_id = self.process_step(step_id)
                        job_ids.append(job_id)
                    except Exception as e:
                        # TODO: cancel running jobs?
                        if is_debug():
                            import traceback

                            traceback.print_stack()
                        log.error(
                            f"Processing of step '{step_id}' from pipeline '{self.pipeline.title}' failed: {e}"
                        )
                        return False
                self._processor.wait_for(*job_ids)
                self._finished_until = idx
        finally:
            self._is_running = False



# -*- coding: utf-8 -*-
import abc
import typing
import uuid
import zmq
from datetime import datetime
from pydantic import BaseSettings
from zmq import Context, Socket

from kiara.data.values import ValueInfo, ValueLineage
from kiara.data.values.value_set import ValueSet
from kiara.exceptions import KiaraProcessingException
from kiara.module import KiaraModule, StepInputs, StepOutputs
from kiara.pipeline import PipelineValuesInfo
from kiara.processing import Job, JobLog, JobStatus
from kiara.utils import is_debug

try:
    from typing import Literal

    from kiara.kiara import Kiara

except Exception:
    from typing_extensions import Literal  # type: ignore


class ProcessorConfig(BaseSettings):

    module_processor_type: Literal["synchronous", "multi-threaded"] = "synchronous"


class ModuleProcessor(abc.ABC):
    @classmethod
    def from_config(
        cls,
        config: typing.Union[None, typing.Mapping[str, typing.Any], ProcessorConfig],
        kiara: typing.Optional["Kiara"] = None,
    ) -> "ModuleProcessor":

        from kiara.processing.parallel import ThreadPoolProcessorConfig
        from kiara.processing.synchronous import SynchronousProcessorConfig

        if not config:
            config = SynchronousProcessorConfig(module_processor_type="synchronous")
        if isinstance(config, typing.Mapping):
            processor_type = config.get("module_processor_type", None)
            if not processor_type:
                raise ValueError("No 'module_processor_type' provided: {config}")
            if processor_type == "synchronous":
                config = SynchronousProcessorConfig(**config)
            elif processor_type == "multi-threaded":
                config = ThreadPoolProcessorConfig()
            else:
                raise ValueError(f"Invalid processor type: {processor_type}")

        if isinstance(config, SynchronousProcessorConfig):
            from kiara.processing.synchronous import SynchronousProcessor

            proc: ModuleProcessor = SynchronousProcessor(
                kiara=kiara, **config.dict(exclude={"module_processor_type"})
            )
        elif isinstance(config, ThreadPoolProcessorConfig):
            from kiara.processing.parallel import ThreadPoolProcessor

            proc = ThreadPoolProcessor(
                kiara=kiara, **config.dict(exclude={"module_processor_type"})
            )

        return proc

    def __init__(
        self,
        zmq_context: typing.Optional[Context] = None,
        kiara: typing.Optional["Kiara"] = None,
    ):

        if kiara is None:
            from kiara.kiara import Kiara

            kiara = Kiara.instance()
        self._kiara: Kiara = kiara

        if zmq_context is None:
            zmq_context = Context.instance()
        self._zmq_context: Context = zmq_context
        self._socket: Socket = self._zmq_context.socket(zmq.PUB)
        self._socket.connect("inproc://kiara_in")
        self._active_jobs: typing.Dict[str, Job] = {}
        self._finished_jobs: typing.Dict[str, Job] = {}

        # TODO: clean up those?
        self._inputs: typing.Dict[str, StepInputs] = {}
        self._outputs: typing.Dict[str, StepOutputs] = {}

    def get_job_details(self, job_id: str) -> typing.Optional[Job]:

        if job_id in self._active_jobs.keys():
            return self._active_jobs[job_id]
        elif job_id in self._finished_jobs.keys():
            return self._finished_jobs[job_id]
        else:
            return None

    def start(
        self,
        pipeline_id: str,
        pipeline_name: str,
        step_id: str,
        module: "KiaraModule",
        inputs: ValueSet,
        outputs: ValueSet,
    ) -> str:

        job_id = str(uuid.uuid4())

        # TODO: make snapshot of current state of data?

        full_inputs = module.create_full_inputs(**inputs)
        wrapped_inputs = StepInputs(inputs=full_inputs, kiara=self._kiara)
        wrapped_outputs = StepOutputs(outputs=outputs, kiara=self._kiara)

        self._inputs[job_id] = wrapped_inputs
        self._outputs[job_id] = wrapped_outputs

        job = Job(
            id=job_id,
            pipeline_id=pipeline_id,
            pipeline_name=pipeline_name,
            step_id=step_id,
            module_type=module._module_type_id,  # type: ignore
            module_config=module.config.dict(),
            module_doc=module.get_type_metadata().documentation,
            inputs=PipelineValuesInfo.from_value_set(inputs),
            outputs=PipelineValuesInfo.from_value_set(outputs),
        )
        job.job_log.add_log("job created")
        self._active_jobs[job_id] = job
        self._socket.send_string(Job.create_event_msg(job))

        try:
            self.process(
                job_id=job_id,
                module=module,
                inputs=wrapped_inputs,
                outputs=wrapped_outputs,
                job_log=job.job_log,
            )

            return job_id
        except Exception as e:
            job.error = str(e)
            if is_debug():
                try:
                    import traceback

                    traceback.print_exc()
                except Exception:
                    pass
            if isinstance(e, KiaraProcessingException):
                e._module = module
                e._inputs = inputs
                raise e
            else:
                raise KiaraProcessingException(e, module=module, inputs=inputs)

    def job_status_updated(
        self, job_id: str, status: typing.Union[JobStatus, str, Exception]
    ):

        job = self._active_jobs.get(job_id, None)
        if job is None:
            raise Exception(
                f"Can't retrieve active job with id '{job_id}', no such job registered."
            )

        if status == JobStatus.SUCCESS:
            job.job_log.add_log("job finished successfully")
            job.status = JobStatus.SUCCESS
            job = self._active_jobs.pop(job_id)
            job.finished = datetime.now()
            self._finished_jobs[job_id] = job
            self._socket.send_string(Job.create_event_msg(job))
        elif status == JobStatus.FAILED or isinstance(status, (str, Exception)):
            job.job_log.add_log("job failed")
            job.status = JobStatus.FAILED
            job.finished = datetime.now()
            if isinstance(status, str):
                job.error = status
            elif isinstance(status, Exception):
                job.error = str(status)
            job = self._active_jobs.pop(job_id)
            self._finished_jobs[job_id] = job
            self._socket.send_string(Job.create_event_msg(job))
        elif status == JobStatus.STARTED:
            job.job_log.add_log("job started")
            job.status = JobStatus.STARTED
            self._socket.send_string(Job.create_event_msg(job))
        else:
            raise ValueError(f"Invalid value for status: {status}")

    def sync_outputs(self, *job_ids: str):

        for j_id in job_ids:
            d = self._outputs[j_id]
            d.sync()

            job_inputs = self._inputs[j_id]
            job_details = self.get_job_details(j_id)
            if not job_details:
                raise Exception(f"Can't sync outputs, no job with id: {j_id}")

            input_infos: typing.Dict[str, ValueInfo] = {
                k: v.get_info() for k, v in job_inputs.items()
            }
            for field_name, value in d.items():
                value_lineage = ValueLineage.create(
                    module_type=job_details.module_type,
                    module_config=job_details.module_config,
                    module_doc=job_details.module_doc,
                    output_name=field_name,
                    inputs=input_infos,
                )
                value.set_value_lineage(value_lineage)

    def wait_for(self, *job_ids: str, sync_outputs: bool = True):
        """Wait for the jobs with the specified ids, also optionally sync their outputs with the pipeline value state."""

        self._wait_for(*job_ids)

        if sync_outputs:
            self.sync_outputs(*job_ids)

    @abc.abstractmethod
    def process(
        self,
        job_id: str,
        module: "KiaraModule",
        inputs: ValueSet,
        outputs: ValueSet,
        job_log: JobLog,
    ) -> str:
        pass

    @abc.abstractmethod
    def _wait_for(self, *job_ids: str):

        pass



# -*- coding: utf-8 -*-
import logging
import typing
from datetime import datetime
from enum import Enum
from pydantic import BaseModel, Field, validator

from kiara.metadata.core_models import DocumentationMetadataModel
from kiara.pipeline import PipelineValuesInfo

try:
    from typing import Literal
except Exception:
    from typing_extensions import Literal  # type: ignore  # noqa


class JobStatus(Enum):

    CREATED = "__job_created__"
    STARTED = "__job_started__"
    SUCCESS = "__job_success__"
    FAILED = "__job_failed__"


class LogMessage(BaseModel):

    timestamp: datetime = Field(
        description="The time the message was logged.", default_factory=datetime.now
    )
    log_level: int = Field(description="The log level.")
    msg: str = Field(description="The log message")


class JobLog(BaseModel):

    log: typing.Dict[int, LogMessage] = Field(
        description="The logs for this job.", default_factory=dict
    )
    percent_finished: int = Field(
        description="Describes how much of the job is finished. A negative number means the module does not support progress tracking.",
        default=-1,
    )

    def add_log(self, msg: str, log_level: int = logging.DEBUG):

        _msg = LogMessage(msg=msg, log_level=log_level)
        self.log[len(self.log)] = _msg


class Job(BaseModel):
    @classmethod
    def create_event_msg(cls, job: "Job"):

        topic = job.status.value[2:-2]

        payload = f"{topic} {job.json()}"
        return payload

    class Config:
        use_enum_values = True

    id: str = Field(description="The id of the job.")
    pipeline_id: str = Field(description="The id of the pipeline this jobs runs for.")
    pipeline_name: str = Field(description="The name/type of the pipeline.")
    step_id: str = Field(description="The id of the step within the pipeline.")
    module_type: str = Field(description="The module type name.")
    module_config: typing.Dict[str, typing.Any] = Field(
        description="The module configuration."
    )
    module_doc: DocumentationMetadataModel = Field(
        description="Documentation for the module that runs the job."
    )
    inputs: PipelineValuesInfo = Field(description="The input values.")
    outputs: PipelineValuesInfo = Field(description="The output values.")
    job_log: JobLog = Field(
        description="Details about the job progress.", default_factory=JobLog
    )
    submitted: datetime = Field(
        description="When the job was submitted.", default_factory=datetime.now
    )
    started: typing.Optional[datetime] = Field(
        description="When the job was started.", default=None
    )
    finished: typing.Optional[datetime] = Field(
        description="When the job was finished.", default=None
    )
    status: JobStatus = Field(
        description="The current status of the job.",
        default=JobStatus.CREATED,
    )
    error: typing.Optional[str] = Field(description="Potential error message.")

    @validator("status")
    def _validate_status(cls, v):

        if isinstance(v, int):
            if v < 0 or v > 100:
                raise ValueError(
                    "Status must be a status string, or an integer between 0 and 100."
                )

        return v



# -*- coding: utf-8 -*-

"""Base module that holds [PipelineModule][kiara.pipeline.module.PipelineModule] classes that are auto-generated
from pipeline descriptions in the ``pipelines`` folder."""
import importlib
import typing

# TODO: add classloader for those classes to runtime
from types import ModuleType

from kiara.defaults import KIARA_MODULE_METADATA_ATTRIBUTE


def create_pipeline_class(
    cls_name: str,
    type_name: str,
    pipeline_desc: typing.Mapping[str, typing.Any],
    base_module: str,
):

    from kiara.pipeline.config import PipelineConfig
    from kiara.pipeline.module import PipelineModule

    pmc = PipelineConfig(**pipeline_desc)

    def init(self, id: str, **kwargs):
        # TODO: merge config
        if kwargs.get("module_config", None):
            raise Exception(
                f"Can't dynamically create PipelineModuleClass, 'module_config' provided externally: {pipeline_desc}"
            )
        kwargs["module_config"] = pipeline_desc
        super(self.__class__, self).__init__(id=id, **kwargs)

    attrs = {
        "__init__": init,
        "__doc__": pmc.documentation,
        "_config_cls": PipelineConfig,
        "_base_pipeline_config": pmc,
    }

    if pmc.context:
        md = dict(pmc.context)
    else:
        md = {}

    md.setdefault("tags", []).append("pipeline")
    md.setdefault("labels", {})["pipeline"] = "yes"
    attrs[KIARA_MODULE_METADATA_ATTRIBUTE] = md

    if not base_module:
        base_module = "kiara.modules.pipelines"
    elif base_module.endswith("pipelines"):
        # TODO: this special case needs documentation
        base_module = base_module.rsplit(".", maxsplit=1)[0]

    # TODO: investiage whether just ignoring non-existing modules is good enough
    _m: typing.Optional[ModuleType] = None
    _base_module: typing.Optional[str] = None
    if "." in type_name:
        intermediate_namespace_tokens = type_name.split(".")[0:-1]
        for token in intermediate_namespace_tokens:
            try:
                _temp = f"{base_module}.{token}"
                _m = importlib.import_module(_temp)
                _base_module = _temp
            except Exception:
                if _m is None:
                    _m = importlib.import_module(base_module)
                if _base_module is None:
                    _base_module = base_module
                break
    else:
        _m = importlib.import_module(base_module)
        _base_module = base_module

    if hasattr(_m, cls_name):
        # TODO: I don't like this, but this is needed for streamlit to work, since it starts a few threads
        # right away, without any possibility to use a lock
        cls = getattr(_m, cls_name)
        return cls
        # raise Exception(
        #     f"Can't attach generated class '{cls_name}' to module '{_base_module}': module already has an attribute with that name."
        # )

    attrs["__module__"] = _base_module

    cls = type(cls_name, (PipelineModule,), attrs)

    setattr(_m, cls_name, cls)

    return cls



# -*- coding: utf-8 -*-
import typing
from pydantic import Field

from kiara.module_config import ModuleConfig


class DeserializeConfig(ModuleConfig):

    # value_id: str = Field(description="The id of the value.")
    serialization_type: str = Field(description="The serialization type.")
    input: typing.Any = Field(
        description="The inputs to use when running this module.", default_factory=dict
    )
    output_name: str = Field(description="The name of the output field for the value.")


class LoadConfig(ModuleConfig):

    value_id: str = Field(description="The id of the value.")
    base_path_input_name: typing.Optional[str] = Field(
        description="The name of the input that stores the base_path where the value is saved.",
        default=None,
    )
    inputs: typing.Dict[str, typing.Any] = Field(
        description="The inputs to use when running this module.", default_factory=dict
    )
    output_name: str = Field(description="The name of the output field for the value.")


class SaveConfig(ModuleConfig):

    inputs: typing.Dict[str, typing.Any] = Field(
        description="The inputs to use when running this module.", default_factory=dict
    )
    load_config_output: str = Field(
        description="The output name that will contain the load config output value."
    )



# -*- coding: utf-8 -*-
import typing
from pydantic import BaseModel, Field, validator

from kiara.module_mgmt import ModuleManager

if typing.TYPE_CHECKING:
    from kiara.module import KiaraModule


try:
    from typing import Literal
except Exception:
    from typing_extensions import Literal  # type: ignore


class PythonModuleManagerConfig(BaseModel):

    module_manager_type: Literal["python"]
    module_classes: typing.Dict[str, typing.Type["KiaraModule"]] = Field(
        description="The module classes this manager should hold."
    )

    @validator("module_classes", pre=True)
    def _ensure_module_class_types(cls, v):

        _classes = []
        if v:
            for _cls in v:
                if isinstance(_cls, str):
                    try:
                        module_name, cls_name = _cls.rsplit(".", maxsplit=1)
                        module = __import__(module_name)
                        _cls = getattr(module, cls_name)
                    except Exception:
                        raise ValueError(
                            f"Can't parse value '{_cls}' into KiaraModule class."
                        )

                from kiara import KiaraModule  # noqa

                if not issubclass(_cls, KiaraModule):
                    raise ValueError(f"Not a KiaraModule sub-class: {_cls}")
                _classes.append(_cls)

        return _classes


class PythonModuleManager(ModuleManager):
    def __init__(
        self,
        module_classes: typing.Optional[
            typing.Mapping[str, typing.Type["KiaraModule"]]
        ] = None,
    ):

        if not module_classes:
            from kiara.utils.class_loading import find_all_kiara_modules

            module_classes = find_all_kiara_modules()

        self._module_classes: typing.Mapping[str, typing.Type[KiaraModule]] = {}

        # TODO: investigate why streamlit reloads this several times
        for k, v in module_classes.items():
            if not v.is_pipeline():
                self._module_classes[k] = v

    def get_module_class(self, module_type: str) -> typing.Type["KiaraModule"]:

        cls = self._module_classes.get(module_type, None)
        if cls is None:
            raise ValueError(f"No module of type '{module_type}' available.")
        return cls

    def get_module_types(self) -> typing.Iterable[str]:
        return self._module_classes.keys()



# -*- coding: utf-8 -*-

from click.testing import CliRunner

from kiara import Kiara
from kiara.interfaces.cli import cli


def test_data_subcommand():

    runner = CliRunner()
    result = runner.invoke(cli, "data")
    assert result.exit_code == 0
    assert "Print the metadata" in result.stdout


def test_data_list_subcommand(presseeded_data_store: Kiara):

    runner = CliRunner()
    result = runner.invoke(
        cli,
        "data list --all",
        env={"KIARA_DATA_STORE": presseeded_data_store.config.data_store},
    )

    assert result.exit_code == 0
    # assert "journal_nodes_table" in result.stdout
    assert "journal_nodes@1" in result.stdout


# async def test_data_explain_subcommand(presseeded_data_store: Kiara):
#
#     runner = CliRunner()
#     result = await runner.invoke(
#         cli,
#         "data explain journal_nodes",
#         env={"KIARA_DATA_STORE": presseeded_data_store.config.data_store},
#     )
#     assert result.exit_code == 0
#
#     assert "table" in result.stdout
#
#
# async def test_data_load_subcommand(presseeded_data_store: Kiara):
#
#     runner = CliRunner()
#     result = await runner.invoke(
#         cli,
#         "data load journal_nodes",
#         env={"KIARA_DATA_STORE": presseeded_data_store.config.data_store},
#     )
#     assert result.exit_code == 0
#     assert "Id" in result.stdout
#     assert "City" in result.stdout



# -*- coding: utf-8 -*-
import click
import sys

from kiara.info.kiara import KiaraInfo
from kiara.utils.output import rich_print


@click.command()
@click.argument("topic", nargs=1, required=False)
@click.pass_context
def info(ctx, topic):
    """kiara context information"""

    kiara_obj = ctx.obj["kiara"]

    if not topic:
        info = KiaraInfo.create(kiara=kiara_obj)
        rich_print(info)
    else:

        if topic not in KiaraInfo.__fields__.keys():
            print(
                f"Info topic '{topic}' not available. Available topics: {', '.join(KiaraInfo.__fields__.keys())}"
            )
            sys.exit(1)

        info = KiaraInfo.create(kiara=kiara_obj)
        rich_print(getattr(info, topic))



# -*- coding: utf-8 -*-
import click

from kiara import Kiara
from kiara.interfaces.rest_api import KiaraRestService


@click.group()
@click.pass_context
def service(ctx):
    """Metadata-related sub-commands."""


@service.command(name="start")
@click.pass_context
async def start(ctx):

    kiara_obj: Kiara = ctx.obj["kiara"]

    service = KiaraRestService(kiara=kiara_obj)
    service.start()

    # app = FastAPI()
    #
    # @app.post("/run")
    # def run():
    #
    #     module_type = "logic.and"
    #     module_config = {"delay": 0.1}
    #     inputs = {"a": True, "b": True}
    #     result_values = kiara_obj.run(
    #         module_type=module_type,
    #         module_config=module_config,
    #         inputs=inputs,
    #         resolve_result=False,
    #     )
    #
    #     op_mgmt: DataOperationMgmt = kiara_obj.data_operations
    #
    #     result = {}
    #     for field_name, value in result_values.items():
    #         v = op_mgmt.run("serialize", "msgpack", value)
    #         result[field_name] = v.get_value_obj("bytes")
    #
    #     r = {}
    #     schemas = {}
    #
    #     for field_name, value in result.items():
    #         assert value.type_name == "bytes"
    #         v = op_mgmt.run("deserialize", "msgpack", value)
    #         r[field_name] = v.get_value_obj("value_data")
    #         schemas[field_name] = {"type": v.get_value_data("value_type")}
    #
    #     value_set = ValueSetImpl.from_schemas(
    #         kiara=kiara_obj, schemas=schemas, initial_values=r
    #     )
    #     import pp
    #
    #     pp(value_set.get_all_value_data())
    #
    # thread = threading.Thread(target=uvicorn.run, args=(app,))
    # thread.start()
    # thread.join()


# def create_module_router_from_modules(kiara: Kiara) -> APIRouter:
#
#     module_router = APIRouter()
#     workflow_router = APIRouter()

# for name in sorted(all_classes):
#
#     cls = all_classes[name]
#
#     if name == "dharpa_workflow":
#         continue
#
#     func, resp_model = create_processing_function(cls)
#
#     is_workflow = True if issubclass(cls, DharpaWorkflowOld) else False
#     if is_workflow:
#         workflow_router.post(f"/{name}", response_model=resp_model)(func)
#     else:
#         module_router.post(f"/{name}", response_model=resp_model)(func)
#
# return module_router, workflow_router



# -*- coding: utf-8 -*-
import typing
from rich import box
from rich.console import RenderableType
from rich.table import Table

from kiara.info import KiaraInfoModel
from kiara.metadata.module_models import KiaraModuleTypeMetadata

if typing.TYPE_CHECKING:
    from kiara.kiara import Kiara
    from kiara.module import KiaraModule


# class ModuleInfo(KiaraInfoModel):
#     """A simple model class to hold and display information about a module.
#
#     This is not used in processing at all, it is really only there to make it easier to communicate module characteristics..
#     """
#
#     @classmethod
#     def from_module_cls(cls, module_cls: typing.Type["KiaraModule"]):
#
#         return ModuleInfo(
#             metadata=KiaraModuleTypeMetadata.from_module_class(module_cls=module_cls)
#         )
#
#     class Config:
#         extra = Extra.forbid
#         allow_mutation = False
#
#     metadata: KiaraModuleTypeMetadata = Field(description="The metadata of the module.")
#
#     def create_renderable(self, **config: typing.Any) -> RenderableType:
#
#         table = self.metadata.create_renderable()
#         return table


class ModuleTypesGroupInfo(KiaraInfoModel):

    __root__: typing.Dict[str, KiaraModuleTypeMetadata]

    @classmethod
    def from_type_names(
        cls,
        kiara: "Kiara",
        type_names: typing.Optional[typing.Iterable[str]] = None,
        ignore_pipeline_modules: bool = False,
        ignore_non_pipeline_modules: bool = False,
        **config: typing.Any
    ):

        if ignore_pipeline_modules and ignore_non_pipeline_modules:
            raise Exception("Can't ignore both pipeline and non-pipeline modules.")

        if type_names is None:
            type_names = kiara.available_module_types

        classes = {}
        for tn in type_names:
            _cls = kiara.get_module_class(tn)
            if ignore_pipeline_modules and _cls.is_pipeline():
                continue
            if ignore_non_pipeline_modules and not _cls.is_pipeline():
                continue
            classes[tn] = KiaraModuleTypeMetadata.from_module_class(_cls)

        return ModuleTypesGroupInfo(__root__=classes)

    @classmethod
    def create_renderable_from_type_names(
        cls,
        kiara: "Kiara",
        type_names: typing.Iterable[str],
        ignore_pipeline_modules: bool = False,
        ignore_non_pipeline_modules: bool = False,
        **config: typing.Any
    ):

        classes = {}
        for tn in type_names:
            _cls = kiara.get_module_class(tn)
            classes[tn] = _cls

        return cls.create_renderable_from_module_type_map(
            module_types=classes,
            ignore_pipeline_modules=ignore_pipeline_modules,
            ignore_non_pipeline_modules=ignore_non_pipeline_modules,
            **config
        )

    @classmethod
    def create_renderable_from_module_type_map(
        cls,
        module_types: typing.Mapping[str, typing.Type["KiaraModule"]],
        ignore_pipeline_modules: bool = False,
        ignore_non_pipeline_modules: bool = False,
        **config: typing.Any
    ):
        """Create a renderable from a map of module classes.

        Render-configuration options:
          - include_full_doc (default: False): include the full documentation, instead of just a one line description
        """

        return cls.create_renderable_from_module_info_map(
            {
                k: KiaraModuleTypeMetadata.from_module_class(v)
                for k, v in module_types.items()
            },
            ignore_pipeline_modules=ignore_pipeline_modules,
            ignore_non_pipeline_modules=ignore_non_pipeline_modules,
            **config
        )

    @classmethod
    def create_renderable_from_module_info_map(
        cls,
        module_types: typing.Mapping[str, KiaraModuleTypeMetadata],
        ignore_pipeline_modules: bool = False,
        ignore_non_pipeline_modules: bool = False,
        **config: typing.Any
    ):
        """Create a renderable from a map of module info wrappers.

        Render-configuration options:
          - include_full_doc (default: False): include the full documentation, instead of just a one line description
        """

        if ignore_pipeline_modules and ignore_non_pipeline_modules:
            raise Exception("Can't ignore both pipeline and non-pipeline modules.")

        if ignore_pipeline_modules:
            module_types = {k: v for k, v in module_types.items() if not v.is_pipeline}
        elif ignore_non_pipeline_modules:
            module_types = {k: v for k, v in module_types.items() if v.is_pipeline}

        show_lines = False
        table = Table(show_header=False, box=box.SIMPLE, show_lines=show_lines)
        table.add_column("name", style="b")
        table.add_column("desc", style="i")

        for name, details in module_types.items():

            if config.get("include_full_doc", False):
                table.add_row(name, details.documentation.full_doc)
            else:
                table.add_row(name, details.documentation.description)

        return table

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        return ModuleTypesGroupInfo.create_renderable_from_module_info_map(
            self.__root__
        )



# -*- coding: utf-8 -*-
import typing

from kiara import Kiara, PipelineController
from kiara.data import ValueSet
from kiara.module_config import ModuleConfig
from kiara.workflow.kiara_workflow import KiaraWorkflow


class WorkflowRenderer(object):
    def __init__(
        self,
        config: typing.Union[ModuleConfig, typing.Mapping[str, typing.Any], str],
        workflow_id: typing.Optional[str] = None,
        module_config: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        controller: typing.Optional[PipelineController] = None,
        kiara: typing.Optional[Kiara] = None,
    ):

        if kiara is None:
            kiara = Kiara.instance()
        self._kiara = kiara
        self._config: typing.Union[
            ModuleConfig, typing.Mapping[str, typing.Any], str
        ] = config
        self._workflow_id: typing.Optional[str] = workflow_id
        self._module_config: typing.Optional[
            typing.Mapping[str, typing.Any]
        ] = module_config
        self._controller: typing.Optional[PipelineController] = controller
        self._workflow: typing.Optional[KiaraWorkflow] = None

    @property
    def workflow(self) -> KiaraWorkflow:
        if self._workflow is None:
            self._workflow = self._kiara.create_workflow(
                config=self._config,
                workflow_id=self._workflow_id,
                module_config=self._module_config,
                controller=self._controller,
            )
        return self._workflow

    @property
    def inputs(self) -> ValueSet:
        return self.workflow.pipeline.inputs

    @property
    def outputs(self) -> ValueSet:
        return self.workflow.pipeline.outputs
